/* XMLToCSVBasic.cpp

This program converts an input XML instance document
of a Row/ColumXX organization into a comma separated
value file.

Program developed in Chapter 2
Using XML with Legacy Business Applications

Source code is Copyright (c) 2002 Addison-Wesley Professional
Author:
Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License,
Version 2, as published by the Free Software Foundation.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA
The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"

//  Program specific header files
#include "CSVRowWriter.h"


int main(int argc, char* argv[])
{

  // Local variables

  char cUsageMessage[] =
	  "Usage: XMLToCSVBasic InputFile.XML OutputFile.CSV options \
or XMLToCSVBasic -h";

  char cHelpMessage[] = 
"Converts an input XML document in Row/ColumnXX \n\
format into an output comma separated value file. \n\
\n\
Arguments:\n\
  InputFile.XML    Input XML instance document\n\
  OutputFile.CSV   Output CSV file\n\
\n\
Options:\n\
  -h   Display this message and exit with no action\n\
\n\
Copyright (c) 2002 Addison-Wesley Professional\n\
Author: Michael C. Rawlins, Rawlins EC Consulting\n\
\n\
XMLToCSVBasic\
 comes with with ABSOLUTELY NO WARRANTY.\n\
This is free software and you are welcome to\n\
redistribute it under certain conditions.\n\
Refer to the GNU General Public License for details.\n";


  char * cInputXMLName;
  char * cOutputCSVName;
  int iExitStatus;
  int iRows;
  CSVRowWriter * RowWriter;
  ofstream OutputCSV;

  //  Homegrown exception values
  const char cCreateInstanceError = 1;
  const char cParseError = 2;
  const char cOutputFileError = 3;

  HRESULT hResult = S_OK;

  // Initialize a few things
  iExitStatus = 0;
  iRows = 0;


  // Parse input arguments from command line
  // IF help option is specified
  //     display help message and exit
  // ENDIF
  if ( ( ( argc == 2 ) && (strcmp(argv[1],"-h") == 0 ) ) ||
       ( ( argc == 4 ) && (strcmp(argv[3],"-h") == 0 ) ) )
  {
    cout << cUsageMessage << endl << endl;
    cout << cHelpMessage << endl;
    return 0;
  }
    
  if ((argc < 3) || (argc > 4) )
  {
    cout << cUsageMessage << endl;
    return 1;
  }

  cInputXMLName = argv[1];
  cOutputCSVName = argv[2];


  //  Set up the COM environment, since MSXML is a
  //  COM Component
  hResult = CoInitialize(NULL);
  if (FAILED(hResult))
  {
    cerr << "Failed to initialize COM environment" << endl;
    return 0;
  }

  //  We'll do our main processing within a try block
  //  We only expect to throw COM, MSXML parse
  //  and I/O exceptions
  try
  {
    //  Declare smart pointers for our MSXML COM Interfaces.
    //  We'll use the Dccument2 class for all of our stuff so that
    //  we can do schema validation.
    //  We do these within the try block so that they when we exit
    //  the program they will be out of scope.
    IXMLDOMDocument2Ptr spDocInput;
    IXMLDOMNodeListPtr spRowList;
    IXMLDOMParseErrorPtr spParseError;

    //  Create the COM DOM Document object
    hResult = spDocInput.CreateInstance(__uuidof(DOMDocument40));

    if FAILED(hResult)
    {
      throw cCreateInstanceError;
    }

    //  Tell it we don't want to load the document asynchronously
    spDocInput->async =  VARIANT_FALSE; 

    //  Load input XML Document (dependent on implementation)
    hResult = spDocInput->load(cInputXMLName);
    
    // Check for errors
    if( hResult != VARIANT_TRUE)
    {
      spParseError = spDocInput->GetparseError();
      cerr << "Parsing Error" << endl;
      displayParseError(spParseError);
      throw cParseError;
    }

  	// Open output file
	  OutputCSV.open(cOutputCSVName, ios::out );
  	if ( ! OutputCSV )
    {
      throw cOutputFileError;
    }
    
    //  Initialize CSVRowWriter object
    RowWriter = new CSVRowWriter(&OutputCSV);

    //  NodeList of Rows <- Call Document's
    //    getElementsByTagName for all elements named Row
    spRowList = spDocInput->getElementsByTagName("Row");
      

    //  DO until Rows NodeList.item[index] is null
    //    Call CSVRowWriter write method, passing
    //      NodeList.item[index]
    //    Increment index
    //  ENDDO

    while (spRowList->item[iRows] != NULL)
    {
      IXMLDOMElementPtr spRow = spRowList->item[iRows];
      RowWriter->write(spRow);
      iRows++;
    }

  } // End of try block

 
  //  General purpose catch block
  catch (char cException)
  {
    switch (cException)
    {
    case  cCreateInstanceError:
      cerr << "Failed to create Document instance" << endl;
      cerr << "HRESULT = x" << hex << hResult << endl;
    case  cParseError:
      break;
    case  cOutputFileError:
      cerr << "Failure in opening output file " << 
        cOutputCSVName << endl;
      break;
    default:
      cerr << "Unexpected char exception " << cException
        << endl;
    }
    iExitStatus = 1;
  }
    
  //  Do some simple error handling for COM exceptions
  catch (_com_error &e)
  {
    handleCOMError(e);
    iExitStatus = 1;
  }

  // Finish up
  if (iExitStatus == 0)
  {
    cout << "Number of CSV rows written: " << iRows << endl;
  }
  else
  {
    cerr << "Abnormal Termination" << endl;
  }

  //  Cleanup
  if (OutputCSV != NULL)
  {
    OutputCSV.close();
  }

  //  It is advisable to clean up C++ objects before
  //  COM uninitialize
  if (RowWriter != NULL)
  {
    delete RowWriter;
  }

  CoUninitialize();

  return iExitStatus;
}
