/* CSVRowReader.cpp

This class takes as input a row read from a CSV file, parses it,
and creates a DOM Element named Row with ColumnXX element
children with Text nodes with the content.

From Chapter 3 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with XMLToCSVBasic; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "CSVRowReader.h"


//  Constructor
CSVRowReader::CSVRowReader(IXMLDOMDocument2Ptr ispDocOutput,
    IXMLDOMElementPtr ispEleRoot)
{
  spDocOutput = ispDocOutput;
  spEleRoot = ispEleRoot;

  //  Initialize the column Array
  for (int i = 0; i < MAXCOLUMNS+1; i++)
  {
    cColumnArray[i] = NULL;
  }

}

//  Destructor
CSVRowReader::~CSVRowReader() {};


//  parse - Separate the input row into columns

void CSVRowReader::parse(char * cInputRecord)
{

  // Local Variables
  int iColumnNumber, iCurrentPosition, iRowLength, 
    iCurrentCell;
  bool boQuoteOpen, boColumnActive;

  //  Initializations
  iColumnNumber = 1;
  iCurrentPosition = 0;
  iRowLength = strlen(cInputRecord);
  iHighestColumn = 0;
  boQuoteOpen = false;
  boColumnActive = false;

  //  DO until CurrentPosition = Length of InputRecord
  while ( (iCurrentPosition < iRowLength ) &&
    ( iColumnNumber <= MAXCOLUMNS ) )
  {
    //  DO CASE of InputRecord[CurrentPosition]
    switch (cInputRecord[iCurrentPosition])
    {
      //  Quote:
      //    IF QuoteOpen is true
      //      QuoteOpen <- false
      //    ELSE
      //      QuoteOpen <- true
      //    ENDIF
      //    BREAK
      case  '\"':
        if (boQuoteOpen)
        {
          boQuoteOpen = false;
        }
        else
        {
          boQuoteOpen = true;
        }
        break;

      //  Comma:
      //    IF QuoteOpen is false
      //      Increment ColumnNumber
      //      ColumnActive <- false
      //      BREAK
      //    ENDIF
      case  ',':
        if (!boQuoteOpen)
        {
          iColumnNumber++;
          boColumnActive = false;
          break;
        }
        //  NOTE:  We fall through to Other for commas within
        //  Quoted strings

      //  Other:
      //    IF ColumnActive is false
      //      ColumnActive <- true
      //      HighestColumn <-ColumnNumber
      //    ENDIF
      //    ColumnArray[ColumnNumber] <-
      //      ColumnArray[ColumnNumber] +
      //      InputRecord[CurrentPosition]
      default:
        if (!boColumnActive)
        {
          boColumnActive = true;
          iHighestColumn = iColumnNumber;
          //  Allocate memory for this column
          cColumnArray[iColumnNumber] =
            new char[MAXCOLUMNSIZE];
          memset(cColumnArray[iColumnNumber],'\0',
            MAXCOLUMNSIZE);
          iCurrentCell = -1;
        }
        iCurrentCell++;
        cColumnArray[iColumnNumber][iCurrentCell] =
          cInputRecord[iCurrentPosition];

    } //  ENDDO CASE

    //  Increment CurrentPosition
    iCurrentPosition++;
  }  //  ENDDO

}  


//  write - Write from the column array to a Row node
//  with ColumnXX children
void CSVRowReader::write()
{
  // Local Variables
  IXMLDOMElementPtr spEleRow;

  char cColumnName[10], cColumnNumber[3];
  int iColumnNumber;

  //  Row Element <- Call Document's createElement method with
  //    "Row" as the tagName
  spEleRow = spDocOutput->createElement("Row");

  //  Root Element <- Call Root Element's appendChild to add
  //    Row Element as child
  spEleRoot->appendChild(spEleRow);

  //  DO for all ColumnNumber from 1 to MaxColumns
  //  Increment ColumnNumber
  for (iColumnNumber=1; iColumnNumber <= iHighestColumn;
    iColumnNumber++)
  {
    //  IF ColumnArray[ColumnNumber] has content
    if (cColumnArray[iColumnNumber] != NULL)
    {
      //  Column Name <- "Column" and ColumnNumber
      strcpy(cColumnName,"Column");
      if (iColumnNumber <= 9)
      {
        strcat(cColumnName,"0");
      }
      _itoa(iColumnNumber,cColumnNumber,10);
      strcat(cColumnName,cColumnNumber);

      //  Column Element <- Call Document's createElement
      //    method with ColumnName as the tagName
      IXMLDOMElementPtr spEleColumn = 
        spDocOutput->createElement(cColumnName);

      //  Row Node <- Call Row Node's appendChild to add Column
      //    Element as child
      spEleRow->appendChild(spEleColumn);

      //  Text Node <- Call Document's createTextNode method
      //    with text from ColumnArray[ColumnNumber]
      //  MSXML createTextNode specifies a VARIANT for the
      //    argument.  Visual C++ is usually happy to convert
      //    a char *, but wanted an explicit cast for this
      //    array element.
      IXMLDOMTextPtr spTexText = spDocOutput->createTextNode(
        (char *) cColumnArray[iColumnNumber]);

      //  Column Element <- Call Column Element's appendChild
      //    to add Text Node as child
      spEleColumn->appendChild(spTexText);

      //  Free the memory for this cell
      delete cColumnArray[iColumnNumber];
      cColumnArray[iColumnNumber] = NULL;

    } //  ENDIF
  } //  ENDDO
  
  return;
}  
  