/* CSVRecordReader.cpp

Handles reading and parsing rows in CSV files and creating XML

From Chapter 7 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Special headers for this module only
#include "CSVRecordReader.h"

//  Parsing constants for this file only
const char NewColumn = 1;
const char StartDelimitedColumn = 2;
const char FinishDelimitedColumn = 3;
const char RegularColumn = 4;
const char DelimitedColumn = 5;
const char ParsingError = 6;

//  Default constructor
CSVRecordReader::CSVRecordReader(
    const IXMLDOMDocument2Ptr spFileDescription)
    : RecordReader(spFileDescription)
{
  //  Locals
  IXMLDOMElementPtr spEleTemp;
  char cValueTemp[3];

  //  RecordTerminator <- Get RecordTerminator value from File
  //      Description DOM Document
  spEleTemp = spFileDescription->getElementsByTagName(
    "RecordTerminator")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")) );

  //  Call setTerminator to set the Record Terminator
  setTerminator(cValueTemp);

  //  ColumnDelimeter <- Get ColumnDelimeter value from File
  //      Description DOM Document
  spEleTemp = spFileDescription->getElementsByTagName(
    "ColumnDelimiter")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")) );

  cColumnDelimiter = setDelimiter(cValueTemp);

  //  TextDelimeter <- Get TextDelimeter value from File
  //      Description DOM Document
  spEleTemp = spFileDescription->getElementsByTagName(
    "TextDelimiter")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")) );
 
  cTextDelimiter = setDelimiter(cValueTemp);
  
}


//  Destructor
CSVRecordReader::~CSVRecordReader() {};

//  Parse the CSV record
int CSVRecordReader::parseRecord(
    const IXMLDOMElementPtr spRowGrammar)
{
  //  locals

  //  Column Number <- 1
  int iCharactersParsed = 0;
  int iReturn;
  char cInputCharacter;
 
  iColumnNumber = 1;

  //  Column Grammar <- Get first ColumnDescription child from
  //      Row Grammar Element
  spColumnGrammars = spRowGrammar->getElementsByTagName(
    "ColumnDescription");

  //  GrammarsIndex <- -1
  iGrammarIndex = -1;
  //  Parsing State <- New Column
  ParsingState = NewColumn;

  //  DO until end of input record or Parsing State is Error
  while ( (iCharactersParsed < iRecordBufferLength) &&
    (ParsingState != ParsingError) )
  {
    //  Input Character <- Next character from input record
    cInputCharacter = RecordBuffer[iCharactersParsed];
    //  DO CASE of Parsing State
    switch (ParsingState)
    {
      //  New Column:
      case NewColumn:
        //  DO CASE of Input Character
        //  Column Delimiter:
        if (cInputCharacter == cColumnDelimiter)
        {
          //  Increment Column Number
          iColumnNumber++;
        }
        else
        {
          //  Text Delimiter:
          if (cInputCharacter == cTextDelimiter)
          {
            //  Parsing State <- Start Delimited Column
            ParsingState = StartDelimitedColumn;
          }
          //  other:
          else
          {
            //   Call saveCharacter
            iReturn = saveCharacter(cInputCharacter);
            if (iReturn != SUCCESS)
            {
              return iReturn;
            }
            //  Parsing State <- Regular Column
            ParsingState = RegularColumn;
          }
        } //  ENDDO
        break;  //  BREAK
      //  Regular Column:
      case RegularColumn:
        //  DO CASE of Input Character
        //  Column Delimiter:
        if (cInputCharacter == cColumnDelimiter)
        {
          //  Parsing State <- New Column
          ParsingState = NewColumn;
          //  Increment Column Number
          iColumnNumber++;
        }
        //  other:
        else
        {
          //   Call saveCharacter
          iReturn = saveCharacter(cInputCharacter);
          if (iReturn != SUCCESS)
          {
            return iReturn;
          }
        } //  ENDDO
        break;  //  BREAK
      //  Start Delimited Column:
      case StartDelimitedColumn:
        //  DO CASE of Input Character
        //  Text Delimiter:
        if (cInputCharacter == cTextDelimiter)
        {
          //  Parsing State <- Finish Delimited Column
          ParsingState = FinishDelimitedColumn;
        }
        //  other:
        else
        {
          //   Call saveCharacter
          iReturn = saveCharacter(cInputCharacter);
          if (iReturn != SUCCESS)
          {
            return iReturn;
          }
          //  Parsing State <- Delimited Column
          ParsingState = DelimitedColumn;
        }  //    ENDDO
        break;  //  BREAK
      //  Delimited Column:
      case DelimitedColumn:
        //  DO CASE of Input Character
        //  Text Delimiter:
        if (cInputCharacter == cTextDelimiter)
        {
          //  Parsing State <- Finish Delimited Column
          ParsingState = FinishDelimitedColumn;
        }
        //  other:
        else
        {
          //   Call saveCharacter
          iReturn = saveCharacter(cInputCharacter);
          if (iReturn != SUCCESS)
          {
            return iReturn;
          }
        } //  ENDDO
        break;  //  BREAK
      //  Finish Delimited Column:
      case FinishDelimitedColumn:
        //  DO CASE of Input Character
        //    Column Delimiter:
        if (cInputCharacter == cColumnDelimiter)
        {
          //  Parsing State <- New Column
          ParsingState = NewColumn;
          //  Increment Column Number
          iColumnNumber++;
        }
        //  other:
        else
        {
          //  Parsing State <- Error
          ParsingState = ParsingError;
        }
        break;  // BREAK
    } //    ENDDO
    iCharactersParsed++;
  } //  ENDDO

  //  IF Parsing State = Error
  //     Return error
  //  ENDIF
  if (ParsingState == ParsingError)
  {
    cerr << CSV_PARSING_ERROR << RecordBuffer << endl;
    return FAILURE;
  }
  //  Return success
  return SUCCESS;
  
}

//  Save a character we find in parsing to the current
//  DataCell object.  Create a new one if needed
int CSVRecordReader::saveCharacter(char cInputCharacter)
{
  IXMLDOMElementPtr spColumnGrammar;
  DataCell * NewCell;
  
  char cGrammarColumnNumber[4];
  int iReturn, iGrammarColumnNumber;


  //  IF Parsing State is not Regular Column or DelimitedColumn
  if (ParsingState < RegularColumn)
  {
    iGrammarIndex++;
    //  Grammar Column Number <- Call Grammar Nodelist
    //  item(GrammarIndex) getAttribute on "FieldNumber", and
    //  convert to integer
    memset(cGrammarColumnNumber,0,sizeof(cGrammarColumnNumber));
    spColumnGrammar = spColumnGrammars->item[iGrammarIndex];

    //  Force error if we have run out of grammar columns
    if (spColumnGrammar == NULL)
    {
      cerr << iColumnNumber << FIELD_NOT_IN_RECORD_GRAMMAR
           << "RowDescription" << endl;
      return FAILURE;
    }

    strcpy(cGrammarColumnNumber,
        _bstr_t(spColumnGrammar->getAttribute("FieldNumber")));
          
    iGrammarColumnNumber = atoi(cGrammarColumnNumber);

    //  DO while Grammar Column Number < Column Number
    while (iGrammarColumnNumber < iColumnNumber)
    {
      //  Increment GrammarIndex
      iGrammarIndex++;
      //  IF Grammar Nodelist item(GrammarIndex) is null
      //    return error
      //  ENDIF
      if (spColumnGrammars->item[iGrammarIndex] == NULL)
      {
        cerr << iColumnNumber << FIELD_NOT_IN_RECORD_GRAMMAR
             << "RowDescription" << endl;
        return FAILURE;
      }
      //  Grammar Column Number <- Call Grammar Nodelist
      //  item(GrammarIndex) getAttribute on "FieldNumber", and
      //  convert to integer
      memset(cGrammarColumnNumber,0,sizeof(cGrammarColumnNumber));
      spColumnGrammar = spColumnGrammars->item[iGrammarIndex];
      strcpy(cGrammarColumnNumber,
          _bstr_t(spColumnGrammar->getAttribute("FieldNumber")));
        
      iGrammarColumnNumber = atoi(cGrammarColumnNumber);
    } //  ENDDO
    //  IF (Grammar Column Number > Column Number)
    //    return error
    //  ENDIF
    if (iGrammarColumnNumber > iColumnNumber)
    {
      cerr << iColumnNumber << FIELD_NOT_IN_RECORD_GRAMMAR
           << "RowDescription" << endl;
      return FAILURE;
    }
    //  DataCell Array[Column Number] <- call RowHandler's
    //    createDataCell method
    iReturn = createDataCell(iColumnNumber, spColumnGrammar,
      NewCell);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
  } //  ENDIF
  //  Call DataCell Array[HighestCell] putByte method to append
  //    Input Character to DataCell buffer
  iReturn = CellArray[iHighestCell]->putByte(cInputCharacter);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  return SUCCESS;
}
