/* CSVRecordWriter.cpp

Handles creating CSV records from XML input

From Chapter 7 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Special headers for this module only
#include "CSVRecordWriter.h"

//  Constructor with output stream
CSVRecordWriter::CSVRecordWriter(
  const IXMLDOMDocument2Ptr spFileDescription,
  ofstream * const OutputFileIn) :
  RecordWriter(spFileDescription, OutputFileIn)
{

  //  Locals
  IXMLDOMElementPtr spEleTemp;
  char cValueTemp[3];

  //  RecordTerminator <- Get RecordTerminator value from File
  //      Description DOM Document
  spEleTemp = spFileDescription->getElementsByTagName(
    "RecordTerminator")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")) );

  //  Call setTerminator to set the Record Terminator
  setTerminator(cValueTemp);

  //  ColumnDelimeter <- Get ColumnDelimeter value from File
  //      Description DOM Document
  spEleTemp = spFileDescription->getElementsByTagName(
    "ColumnDelimiter")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")) );

  cColumnDelimiter = setDelimiter(cValueTemp);

  //  TextDelimeter <- Get TextDelimeter value from File
  //      Description DOM Document
  spEleTemp = spFileDescription->getElementsByTagName(
    "TextDelimiter")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")) );
 
  cTextDelimiter = setDelimiter(cValueTemp);
}

//  Destructor
CSVRecordWriter::~CSVRecordWriter() {};

//  Write record to CSV file
//  NOTE:  We don't use the grammar Element here, but include
//  it to keep the function signature the same as the base
//  class
int CSVRecordWriter::writeRecord(IXMLDOMElementPtr spRowGrammar)
{

  //  Locals
  int i, iCellColumnNumber, iOutputColumnNumber, iLen, iReturn;
  bool boDelimitText;
  char cColumnContents[CELL_BUFFER_SIZE];

  //  Initialize Output Record Buffer
  memset(RecordBuffer,0,sizeof(RecordBuffer));
  iRecordBufferLength = 0;

  //  Output Column Number = 1
  iOutputColumnNumber = 1;
  //  DO for 0 through Highest Cell
  for (i=0; i <= iHighestCell; i++)
  {
    //  Cell Column Number = Get Cell's Field Number
    iCellColumnNumber = CellArray[i]->getFieldNumber();
    //  DO while Output Column Number < Cell Column Number
    while (iOutputColumnNumber < iCellColumnNumber)
    {
      //  Append Column Delimiter to Output Record Buffer
      RecordBuffer[iRecordBufferLength] = cColumnDelimiter;
      iRecordBufferLength++;
      //  Increment Output Column Number
      iOutputColumnNumber++;
    } //  ENDIF

    // Call fromXML to convert Cell contents from XML datatype
    iReturn = CellArray[i]->fromXML();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Delimit Text Flag <- call Cell's delimitText method
    boDelimitText = CellArray[i]->delimitText();

    //  Cell Contents <- Call getField to retrieve cell buffer
    //    contents
    memset(cColumnContents,0,sizeof(cColumnContents));

    iLen = CellArray[i]->getField(cColumnContents);

    //  Clear Array entry
    delete CellArray[i];

    //  IF Delimit Text Flag is true
    if (boDelimitText)
    {
      //  Append Text Delimiter character to Output Record
      //    Buffer
      RecordBuffer[iRecordBufferLength] = cTextDelimiter;
      iRecordBufferLength++;
      //  Append Cell Contents to Output Record Buffer
      memcpy(&(RecordBuffer[iRecordBufferLength]),
          cColumnContents,iLen);
      iRecordBufferLength = iRecordBufferLength + iLen;
      //  Append Text Delimiter character to Output Record
      //    Buffer
      RecordBuffer[iRecordBufferLength] = cTextDelimiter;
      iRecordBufferLength++;
    }
    else  //  ELSE
    {
      //  Append Cell Contents to Output Record Buffer
      memcpy(&(RecordBuffer[iRecordBufferLength]),
          cColumnContents,iLen);
      iRecordBufferLength = iRecordBufferLength + iLen;
    } //  ENDIF

  }   // ENDDO

  iHighestCell = -1;

  //  Append base RecordHandler's Record Terminators to Output
  //    Record Buffer
  RecordBuffer[iRecordBufferLength] = cRecordTerminator1;
  iRecordBufferLength++;
  if ( cRecordTerminator2 != 0 )
  {
    RecordBuffer[iRecordBufferLength] = cRecordTerminator2;
    iRecordBufferLength++;
  }

  //  Call language's write routines to do physical write of
  //    Output Record Buffer
  OutputFile->write(RecordBuffer,iRecordBufferLength);

  if (OutputFile->good())
  {
    return SUCCESS;
  }
  else
  {
    return FAILURE;
  }

}
