/*  CSVSourceConverter.cpp

Drives conversion from CSV to XML

From Chapter 7 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include <direct.h>
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"


//  Program specific header files
#include "CSVRecordReader.h"
#include "CSVSourceConverter.h"
#include <strstrea.h>


//  Default Constructor
CSVSourceConverter::CSVSourceConverter(bool boValidateIn, 
    const char * const cBaseOutputDirectory,
    const char * const cFileDescriptionDocumentName)
    : SourceConverter(boValidateIn, cBaseOutputDirectory)
{
#include "BlasterExceptions.h"
  //  Locals
  IXMLDOMElementPtr spEleTemp;
  IXMLDOMNodeListPtr spNLTemp;
  char cValueTemp[16];

  int i, iReturn;

  //  Initialize Partner Array
  for (i=0; i < MAX_PARTNERS; i++)
  {
    memset(Partners[i],0,sizeof(Partners[i]));
  }
  iNumPartners = 0;

  //  Call loadFileDescriptionDocument from passed document
  iReturn = 
    loadFileDescriptionDocument(cFileDescriptionDocumentName);
  if (iReturn == FAILURE)
  {
    throw cConstructorError;
  }

  //  Schema Location URL <- Call File Description Document's
  //    getElementByTagName on "SchemaLocationURL", then
  //    getAttribute on "value"
  //  NOTE - This is an optional Element, so we may not
  //  get anything
  memset(cSchemaLocationURL,0,sizeof(cSchemaLocationURL));
  spNLTemp = spFileDescription->getElementsByTagName(
      "SchemaLocationURL");
  if (spNLTemp->length != 0)
  {
    spEleTemp = spNLTemp->item[0];
    strcpy(cSchemaLocationURL,
      _bstr_t(spEleTemp->getAttribute("value")));
  }

  //  Throw exception if no schema and validation requested
  if ( boValidate && (strlen(cSchemaLocationURL) == 0) )
  {
    cerr << VALIDATION_WITHOUT_SCHEMA << endl;
    throw cConstructorError;
  }

  //  DocumentBreakColumn <- Get DocumentBreakColumn value from
  //      File Description DOM Document
  memset(cValueTemp,0,sizeof(cValueTemp));
  spEleTemp = (IXMLDOMElementPtr)
    spFileDescription->getElementsByTagName(
    "DocumentBreakColumn")->item[0];
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")));
  iDocumentBreakColumn = atoi(cValueTemp);

  //  PartnerBreakColumn <- Get PartnerBreakColumn value from
  //      File Description DOM Document
  memset(cValueTemp,0,sizeof(cValueTemp));
  spEleTemp = (IXMLDOMElementPtr)
    spFileDescription->getElementsByTagName(
    "PartnerBreakColumn")->item[0];
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")));
  iPartnerBreakColumn = atoi(cValueTemp);


  //  Initialize Saved Partner
  memset(cSavedPartnerID,0,sizeof(cSavedPartnerID));

  //  Initialize Saved DocumentID
  memset(cSavedDocumentID,0,sizeof(cSavedDocumentID));

  //  Create CSVRecordReader object, passing:
  //    File Description Document DOM
  CSVReader = new CSVRecordReader(spFileDescription);
  Reader = CSVReader;
}


//  Destructor
CSVSourceConverter::~CSVSourceConverter() {};

//  Save the output document
int CSVSourceConverter::processFile(
    const char * const cInputFileName)
{
  //  Locals to method

  //  COM
  HRESULT hResult;
  
  //  XML
  IXMLDOMDocument2Ptr spDocOutput;
  IXMLDOMElementPtr spRootElement;
  IXMLDOMElementPtr spRowGrammar;
  IXMLDOMProcessingInstructionPtr spProlog;
  IXMLDOMAttributePtr spSchemaLocationAttribute;

  //  File input and output streams
  ifstream InputFile;

  //  Basic types
  char cOutputDirectory[MAX_PATH_LENGTH];
  char cOutputFilePath[MAX_PATH_LENGTH];
  char cSequenceNumber[4];
  char cSequenceTemp[4];
  int iSequenceNumber, iRecordLength, i, iReturn;
  bool boPartnerFound, boPartnerBreak, boDocumentBreak;

  //  Initializations
  //  Output Directory Path <- Base Directory + directory
  //    separator
  memset(cOutputDirectory,0,sizeof(cOutputDirectory));
  strcpy(cOutputDirectory,cBaseOutputDirectory);
  memset(cOutputFilePath,0,sizeof(cOutputFilePath));
  spDocOutput = NULL;
  spRootElement = NULL;

  //Initialize Sequence Number
  iSequenceNumber = 0;

  //  Row Grammar <- Get RowDescription from Grammar Element
  spRowGrammar =
    spGrammar->getElementsByTagName("RowDescription")->item[0];

  //  Open input file
  InputFile.open(cInputFileName, ios::in | ios::nocreate);
  if ( ! InputFile )
  {
    cerr << INPUT_FILE_OPEN_ERROR << cInputFileName << endl;
    return FAILURE;
  }

    //  C++ Only - set mode to binary
  InputFile.setmode(filebuf::binary);


  //  Call CSVRecordReader's setInputStream method
  CSVReader->setInputStream(&InputFile);

  //  Display processing message
  cout << CREATING_DOCUMENT << endl;
  
  //  Call CSVRecordReader's readRecord method
  iRecordLength = CSVReader->readRecordVariableLength();

  //  DO until end of input file
  while (iRecordLength >= 0)
  {
    //  Call CSVRecordReader's parseRecord method to parse
    //    columns into DataCell array
    iReturn = CSVReader->parseRecord(spRowGrammar);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Call CSVRecordReader's toXMLType method to convert
    //    column contents in DataCell array to their XML Schema
    //    datatype representation
    iReturn = CSVReader->toXMLType();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  PartnerBreak <- Call testPartnerBreak
    //  IF PartnerBreak = true
    iReturn = testPartnerBreak(boPartnerBreak);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
    if ( boPartnerBreak)
    {
      //  Output Directory Path <- Base Directory +
      //    Partner ID + directory separator
      memset(cOutputDirectory,0,sizeof(cOutputDirectory));
      strcpy(cOutputDirectory,cBaseOutputDirectory);
      strcat(cOutputDirectory, cSavedPartnerID);
      strcat(cOutputDirectory, "\\");

      //  Lookup Partner in Partner Array
      boPartnerFound = false;
      for(i = 0; i < iNumPartners; i++)
      {
        if (!strcmp(Partners[i],cSavedPartnerID))
        {
          boPartnerFound = true;
          break;
        }
      }

      //  IF Partner is not in Array
      if (!boPartnerFound)
      {
        //  Partner Array <- Add Partner ID
        strcpy(Partners[iNumPartners],cSavedPartnerID);
        iNumPartners++;

        //  Create output directory from Output Directory
        //  Path
        iReturn = _mkdir(cOutputDirectory);
        if (iReturn != SUCCESS)
        {
          cerr << IO_DIRECTORY_CREATE_ERROR_MESSAGE
               << cOutputDirectory << endl;
          return FAILURE;
        }
      } //  ENDIF

    } //  ENDIF

    //  DocumentBreak <- Call testDocumentBreak
    iReturn = testDocumentBreak(boDocumentBreak);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
    //  IF DocumentBreak = true or Output DOM Document is null
    if ( boDocumentBreak || (spDocOutput == NULL) )
    {
      //    IF Output DOM document exists
      //      Call saveDocument
      //    ENDIF
      if ( spDocOutput != NULL)
      {
        iReturn = saveDocument(spDocOutput,cOutputFilePath,
              boValidate);
        if (iReturn != SUCCESS)
        {
          return FAILURE;
        }
        spDocOutput = NULL;
      }

      //  Create new Output DOM Document
      hResult = spDocOutput.CreateInstance(
        __uuidof(DOMDocument40));

      if FAILED(hResult)
      {
        cerr << CREATE_INSTANCE_ERROR << endl;
        return FAILURE;
      }

      //  Create the header processing instruction
      //    and append it to the document node
      spProlog = spDocOutput->createProcessingInstruction("xml",
        XML_PROLOG_TEXT);
      spDocOutput->appendChild(spProlog);

      //  Create Root Element and append to Output DOM Document
      spRootElement = 
        spDocOutput->createElement(cRootElementName);

      spDocOutput->appendChild(spRootElement);

      //  IF Schema Location URL is not NULL
      if (strlen(cSchemaLocationURL) > 0 )
      {
        //  Create noNameSpaceSchemaLocation Attribute and
        //    append to Root Element
        _variant_t varType((short)NODE_ATTRIBUTE);
        spSchemaLocationAttribute = spDocOutput->createNode(
          varType,"xsi:noNamespaceSchemaLocation",
          "http://www.w3.org/2001/XMLSchema-instance");

        spRootElement->setAttributeNode(spSchemaLocationAttribute);
        // We can finally set it now

        spRootElement->setAttribute("xsi:noNamespaceSchemaLocation",
          cSchemaLocationURL);
      }  //    ENDIF

      //  Error if greater than maximum
      if ((iSequenceNumber+1) > MAX_SEQUENCE_NUMBER)
      {
        cerr << MAX_SEQUENCE_EXCEEDED
             << iSequenceNumber + 1 << endl;
        return FAILURE;
      }
          
      //  Increment Sequence Number
      iSequenceNumber++;

      //  Ouput File Path <- Output directory path +
      //  Root Element Name + Sequence Number + ".xml"
      memset(cSequenceNumber,0,sizeof(cSequenceNumber));
      if ( iSequenceNumber <= 9 )
      {
        strcpy(cSequenceNumber,"00");
        _itoa(iSequenceNumber,cSequenceTemp,10);
        strcat(cSequenceNumber,cSequenceTemp);
      }
      else
      {
        if (iSequenceNumber <= 99)
          {
            strcpy(cSequenceNumber,"0");
            _itoa(iSequenceNumber,cSequenceTemp,10);
            strcat(cSequenceNumber,cSequenceTemp);
          }
          else
          {
            _itoa(iSequenceNumber,cSequenceNumber,10);
          }
      }

      //  Omit sequence number if there is no break on
      //    document
      if (iDocumentBreakColumn != 0)
      {
        memset(cOutputFilePath,0,sizeof(cOutputFilePath));
        strcpy(cOutputFilePath, cOutputDirectory);
        strcat(cOutputFilePath,cRootElementName);
        strcat(cOutputFilePath,cSequenceNumber);
        strcat(cOutputFilePath,".xml");
      }
      else
      {
        memset(cOutputFilePath,0,sizeof(cOutputFilePath));
        strcpy(cOutputFilePath, cOutputDirectory);
        strcat(cOutputFilePath,cRootElementName);
        strcat(cOutputFilePath,".xml");
      }

      //  Display document name
      cout << cOutputFilePath << endl;
      
      //  Call CSVRecordReader's setOutputDocument method for
      //    new document
      CSVReader->setOutputDocument(spDocOutput);
    } //  ENDIF

    //  Call CSVRecordReader writeRecord method to write row
    //    and column Elements from CSVRecordReader's DataCall
    //    Array
    CSVReader->writeRecord(spRootElement, spRowGrammar);

    //  Call CSVRecordReader's readRecord method
    iRecordLength = CSVReader->readRecordVariableLength();
  } //  ENDDO

  //  IF Output DOM document exists
  //    Call saveDocument
  //  ENDIF
  if ( spDocOutput != NULL)
  {
    iReturn = saveDocument(spDocOutput,cOutputFilePath,
        boValidate);
    if (iReturn != SUCCESS)
    {
      return FAILURE;
    }
    spDocOutput = NULL;
  }

  // Close input file
  InputFile.close();

  //  Clear Partners Array
  for(i = 0; i < iNumPartners; i++)
  {
    memset(Partners[i],0,sizeof(Partners[i]));
  }

  iNumPartners = 0;

  // Display completion message
  cout << iSequenceNumber << DOCUMENTS_PROCESSED << endl;

  return SUCCESS;
}

int CSVSourceConverter::testDocumentBreak(bool &boDocumentBreak)
{
  //  Local
  int iReturn;
  char cDocumentID[CELL_BUFFER_SIZE];
  memset(cDocumentID,0,sizeof(cDocumentID));

  //  IF Document Break Column is zero
  //    return false
  //  ENDIF
  if (iDocumentBreakColumn == 0)
  {
    boDocumentBreak = false;
    return SUCCESS;
  }

  //  DocumentID <- Call CSVRecordReader's getFieldValue method
  //    passing the Document Break Column number
  iReturn = CSVReader->getFieldValue(iDocumentBreakColumn,
      cDocumentID);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  IF DocumentID = Saved Document ID
  if (!(strcmp(cDocumentID,cSavedDocumentID)))
  {
    //  return false
    boDocumentBreak = false;
    return SUCCESS;
  } //  ENDIF

  //  Saved Document ID <- Document ID
  memset(cSavedDocumentID,0,sizeof(cSavedDocumentID));
  strcpy(cSavedDocumentID,cDocumentID);

  //  Return true
  boDocumentBreak = true;
  return SUCCESS;

}

int CSVSourceConverter::testPartnerBreak(bool &boPartnerBreak)
{
  //  Local
  int iReturn;
  char cPartnerID[CELL_BUFFER_SIZE];
  memset(cPartnerID,0,sizeof(cPartnerID));

  //  IF Partner Break Column is zero or Document Break Column
  //        is zero
  //    return false
  //  ENDIF
  if ( (iPartnerBreakColumn == 0) ||
    (iDocumentBreakColumn == 0) )
  {
    boPartnerBreak = false;
    return SUCCESS;
  }

  //  PartnerID <- Call CSVRecordReader's getFieldValue method
  //    passing the Partner Break Column number
  iReturn = CSVReader->getFieldValue(iPartnerBreakColumn,
    cPartnerID);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  IF DocumentID = Saved Document ID
  if (!strcmp(cPartnerID,cSavedPartnerID))
  {
    //  return false
    boPartnerBreak = false;
    return SUCCESS;
  } //  ENDIF

  //  Saved Partner ID <- Partner ID
  memset(cSavedPartnerID,0,sizeof(cSavedPartnerID));
  strcpy(cSavedPartnerID,cPartnerID);

  //  Return true
  boPartnerBreak = true;
  return SUCCESS;
}

