/* TargetConverter.cpp

Drives conversion from XML to CSV

From Chapter 7 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"


//  Special headers for this module only
#include "CSVRecordWriter.h"
#include "CSVTargetConverter.h"
#include <strstrea.h>


//  Default Constructor
CSVTargetConverter::CSVTargetConverter(
     const char * const cFileDescriptionDocumentName,
     ofstream * const OutputStream) : 
     TargetConverter(cFileDescriptionDocumentName)
{
  //  call base class constructor, passing
  //    File Description Document Name

  //  Create CSVRecordWriter object, passing:
  //    File Description Document DOM and output stream
  Writer = new CSVRecordWriter(spFileDescription,
      OutputStream);
}


//  Destructor
CSVTargetConverter::~CSVTargetConverter()
{
  delete Writer;
}

//  Process a Document
int CSVTargetConverter::processDocument(
    const IXMLDOMDocument2Ptr spDocInput)
{
  //  Locals
  IXMLDOMElementPtr spRowGrammar;
  IXMLDOMElementPtr spRootElement;
  IXMLDOMElementPtr spRowElement;
  IXMLDOMNodePtr spChildNode;

  char cGrammarRowName[NMTOKEN_SIZE];
  int iReturn;
 
  //  RowGrammar <- Get RowDescription Element from Grammar
  //    Element
  spRowGrammar =
    spGrammar->getElementsByTagName("RowDescription")->item[0];

  //  Grammar Row Element Name <- get Row Grammar Element's
  //    "ElementName" Attribute
  memset(cGrammarRowName,0,sizeof(cGrammarRowName));
  strcpy(cGrammarRowName,
      _bstr_t(spRowGrammar->getAttribute("ElementName")));

  //  Root Element <- Get Document's documentElement attribute
  spRootElement = spDocInput->documentElement;

  //  Root Element Name <- Get Root Element's tagName attribute
  _bstr_t bstrInputRootElementName(spRootElement->tagName);

  //  IF Root Element Name != Grammar Root Element Name
  if ( !(strcmp(cRootElementName,
         (char *)bstrInputRootElementName) == 0) )
  {
    //  Return error
    cerr << bstrInputRootElementName << ROOT_MISMATCH 
         << cRootElementName << endl;
    return FAILURE;
  } //  ENDIF

  //  Get first child element
  spChildNode = spRootElement->firstChild;

  //  DO for child Row Elements of Root Element
  while (spChildNode != NULL)
  {
    //  Skip over non-Element nodes
    while ( (spChildNode != NULL) &&
       (spChildNode->nodeType != NODE_ELEMENT) )
    {
      spChildNode = spChildNode->nextSibling;
    }
    //  Break at end of list
    if (spChildNode == NULL)
    {
      return SUCCESS;
    }

    spRowElement = spChildNode;

    //  Row Element Name <- Get Row Element's tagName attribute
    _bstr_t bstrRowElementName(spRowElement->tagName);
    
    //  IF (Row Element Name != Grammar Row Element Name)
    //    Return error
    //  ENDIF
    if (strcmp(cGrammarRowName,bstrRowElementName))
    {
      cerr << NOT_IN_DOCUMENT_GRAMMAR << bstrRowElementName 
           << endl;
      return FAILURE;
    }
    //  Call CSVRecordWriter's parseRecord method to load
    //    DataCell Array from XML Elements for Row and Columns
    iReturn = Writer->parseRecord(spRowElement,spRowGrammar);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Call CSVRecordWriter's writeRecord to write output
    //    record from DataCell Array
    iReturn = Writer->writeRecord(spRowGrammar);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Get next Row
    spChildNode = spChildNode->nextSibling;
  } //  ENDDO

  return SUCCESS;
}
