/* XMLToCSV.cpp

This program converts one or more XML instance documents into
a comma separated value file

Program developed in Chapter 7
Using XML with Legacy Business Applications

Source code is Copyright (c) 2002 Addison-Wesley Professional
Author:
Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License,
Version 2, as published by the Free Software Foundation.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA
The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"

//  Program specific header files
#include "CSVRecordWriter.h"
#include "CSVTargetConverter.h"


int main(int argc, char* argv[])
{
#include "BlasterExceptions.h"
  // Local variables


  //  COM & MS
  HRESULT hResult;
  HANDLE hFind;
  WIN32_FIND_DATA FindFileData;

  //  Project Classes

  //  File input and output streams
  ofstream OutputStream;

  //  Basic types
  char cInputDirectoryName[MAX_PATH_LENGTH];
  char cInputDirectoryFindPath[MAX_PATH_LENGTH];
  char cInputFileName[MAX_PATH_LENGTH];
  char * cOutputFileName;
  char * cFileDescriptionDocument;
  char cUsageMessage[256];
  char cHelpMessage[1024];
  bool boValidate;
  int iExitStatus;
  int iDocumentsProcessed;
  int i;
  int iReturn;

  //  Initializations
  memset(cUsageMessage,0,sizeof(cUsageMessage));
  strcpy(cUsageMessage,XML_TO_CSV_USAGE);

  memset(cHelpMessage,0,sizeof(cHelpMessage));
  strcpy(cHelpMessage,XML_TO_CSV_USAGE);
  strcat(cHelpMessage,XML_TO_CSV_HELP);
  strcat(cHelpMessage,HELP_INFO_1);
  strcat(cHelpMessage,"XMLToCSV");
  strcat(cHelpMessage,HELP_INFO_2);

  iExitStatus = SUCCESS;
  boValidate = false;
  iDocumentsProcessed = 0;

  memset(cInputDirectoryName,0,MAX_PATH_LENGTH);
  memset(cInputDirectoryFindPath,0,MAX_PATH_LENGTH);


  // Parse input arguments from command line
  // IF help option is specified
  //     display help message and exit
  // ENDIF
  if ( ( argc == 2 ) && (strcmp(argv[1],"-h") == 0 ) )
  {
    cout << cHelpMessage << endl << endl;
    return SUCCESS;
  }
    
  if ((argc < 4) || (argc > 7) )
  {
    cout << cUsageMessage << endl;
    return FAILURE;
  }

  // Check for validation and help message
  for (i = 4; i < argc; i++)
  {
    if (argv[i][0] == '-')
    {
      switch (argv[i][1])
      {
        case 'h':
          cout << cHelpMessage << endl;
          return SUCCESS;
        case 'v':
          boValidate = true;
          break;
        default:
          cout << cUsageMessage << endl;
          return FAILURE;
      }
    }
    else
    {
      cout << cUsageMessage << endl;
      return FAILURE;
    }
  }

  if (strlen(argv[1]) > (MAX_PATH_LENGTH - 2))
  {
    cerr << BUFFER_OVERFLOW_IN << "main" << endl;
    return FAILURE;
  }

  strcpy(cInputDirectoryName,argv[1]);
  cOutputFileName = argv[2];
  cFileDescriptionDocument = argv[3];

  //  Set up the COM environment, since MSXML is a
  //  COM Component
  hResult = CoInitialize(NULL);
  if (FAILED(hResult))
  {
    cerr << COM_INIT_ERROR << endl;
    return FAILURE;
  }

  //  We'll do our main processing within a try block
  //  We only expect to throw COM, MSXML parse
  //  and I/O exceptions
  try
  {
    //  Declare smart pointers for our MSXML COM Interfaces.
    IXMLDOMDocument2Ptr spDocInput;
    IXMLDOMParseErrorPtr spParseError;

    //  Open output file
    OutputStream.open(cOutputFileName, ios::out );
  	if ( ! OutputStream )
    {
      throw cOutputFileError;
    }

    //  C++ Only - set mode to binary
    OutputStream.setmode(filebuf::binary);

    //  Create new CSVTargetConverter object, passing:
    //    File Description Document Name
    //    Output Stream
    CSVTargetConverter Converter(cFileDescriptionDocument,
        &OutputStream);

    //  MSXML specific - set up environment for loading,
    //  parsing, and validating the input files
    //  Set up DOM XML environment
    hResult = spDocInput.CreateInstance(__uuidof(DOMDocument40));

    if FAILED(hResult)
    {
      throw cCreateInstanceError;
    }

    //  Tell it we don't want to load the document asynchronously
    spDocInput->async =  VARIANT_FALSE; 

    // Tell it not to validate during parsing, if specified
    if (!boValidate)
    {
      spDocInput->validateOnParse = VARIANT_FALSE;
    }

    //  Append slash to input directory specification if
    //  needed
    i = strlen(cInputDirectoryName);
    if (cInputDirectoryName[i-1] != '\\')
    {
      cInputDirectoryName[i] = '\\';
      i++;
    }

    strcpy(cInputDirectoryFindPath,cInputDirectoryName);
    //  For FindFirstFile, append wildcards
    if ( strcmp( &(cInputDirectoryFindPath[i-3]),"*.*") != 0)
    {
      strcat(cInputDirectoryFindPath,"*.*");
    }


    //  Open input directory
    //  Get first file from input directory
    hFind = FindFirstFile(cInputDirectoryFindPath, 
        &FindFileData);
    if (hFind == INVALID_HANDLE_VALUE)
    {
      throw cInputDirectoryOpenError;
    }

    //  Display processing message
    cout << PROCESSING_DOCUMENT << endl;

    //  DO for all files in input directory
    do
    {
      //  Get the full path name
      //  Do we need to do this ???
      //  iLen = GetFullPathName(FindFileData.cFileName, sizeof(szTemp), szTemp, &szRep);
      
      //  Skip directory file paths
      if (FindFileData.cFileName[0] == '.')
      {
        continue;
      }

      //  Build input file name
      memset(cInputFileName,0,MAX_PATH_LENGTH);
      strcpy(cInputFileName,cInputDirectoryName);
      strcat(cInputFileName,FindFileData.cFileName);
      
      //  DOM Document <- Load input file, validating it if
      //    requested
      //  Load input XML Document (dependent on implementation)
      hResult = spDocInput->load(cInputFileName);
    
      // Check for errors
      if( hResult != VARIANT_TRUE)
      {
        spParseError = spDocInput->GetparseError();
        cerr << "Parsing Error" << endl;
        displayParseError(spParseError);
        throw cXMLParseError;
      }

      //  Validate the input document
      if (boValidate)
      {
        spParseError = spDocInput->validate();
        if( spParseError->errorCode != S_OK)
        {
          cerr << "Validation Error" << endl;
          displayParseError(spParseError);
          throw cXMLValidationError;
        }
      }

      //  Display document name being processed
      cout << cInputFileName << endl;

      //  Call CSVTargetConverter processDocument method,
      //    passing the input XML DOM Document
      iReturn = Converter.processDocument(spDocInput);
      if (iReturn != SUCCESS)
      {
        throw cProcessDocumentError;
      }
        
      iDocumentsProcessed++;

    } while (FindNextFile(hFind, &FindFileData) == TRUE);
    //  ENDDO

    //  Close the input directory
    FindClose(hFind);

    //  Close output file
    OutputStream.flush();
    OutputStream.close();
  } //  End of try block
    
 
  //  General purpose catch block
  catch (char cException)
  {
    switch (cException)
    {
    case cCreateInstanceError:
      cerr << CREATE_INSTANCE_ERROR << endl;
      cerr << "HRESULT = x" << hex << hResult << endl;
      break;
    case cOutputFileError:
      cerr << IO_OUTPUT_OPEN_ERROR_MESSAGE 
           << cOutputFileName << endl;
      break;
    case cInputDirectoryOpenError:
      cerr << IO_DIRECTORY_OPEN_ERROR_MESSAGE 
           << cInputDirectoryName << endl;
      break;
    case cXMLParseError:
    case cXMLValidationError:
    case cConstructorError:
    case cProcessDocumentError:
      break;
    default:
      cerr << UNEXPECTED_EXCEPTION << cException
           << endl;
    }
    iExitStatus = FAILURE;
  }
    
  //  Do some simple error handling for COM exceptions
  catch (_com_error &e)
  {
    handleCOMError(e);
    iExitStatus = FAILURE;
  }

  // Close COM Environment
  CoUninitialize();

  //  Print Completion Message
  cout << iDocumentsProcessed << DOCUMENTS_PROCESSED << endl;

  // Finish up
  if (iExitStatus == SUCCESS)
  {
    cout << TERMINATION_NORMAL << endl;
  }
  else
  {
    cout << ABEND << endl;
  }

  return iExitStatus;
}
