/* FlatRecordReader.cpp

Handles reading and parsing records in flat files and creating
XML

From Chapter 8 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Special headers for this module only
#include "FlatRecordReader.h"

//  Default constructor
FlatRecordReader::FlatRecordReader(
    const IXMLDOMDocument2Ptr spFileDescription)
    : RecordReader(spFileDescription)
{
  //  Locals
  IXMLDOMElementPtr spRecordFormat;
  IXMLDOMElementPtr spTagInfo;
  IXMLDOMElementPtr spTempElement;
  IXMLDOMNodePtr spTempNode;
  IXMLDOMNodeListPtr spTempNodeList;
  char cValueTemp[3];

  //  Record Format Element <- Get RecordFormat Element from
  //    File Description DOM Document
  spRecordFormat = spFileDescription->getElementsByTagName(
      "RecordFormat")->item[0];

  //  Child Element <- Get first childNode from Record Format Element,
  //    advancing over non-Element Nodes
  spTempNode = spRecordFormat->firstChild;
  while (spTempNode->nodeType != NODE_ELEMENT)
  {
    spTempNode = spTempNode->nextSibling;
  }
  spTempElement = spTempNode;

  //  Fixed Record Length <- 0
  iFixedRecordLength = 0;

  memset(cValueTemp,0,sizeof(cValueTemp));
  //  IF Child Element NodeName = Fixed
  if (!strcmp(spTempElement->nodeName,"Fixed"))
  {
    //  Fixed Record Length <- Call Child Element's
    //    getAttribute for Length
    strcpy(cValueTemp,
        _bstr_t(spTempElement->getAttribute("Length")) );
    iFixedRecordLength = atoi(cValueTemp);
  }
  else  //  ELSE
  {
    //  Record Terminator <- Call Child Element's getAttribute
    //    for RecordTerminator
    strcpy(cValueTemp,
    _bstr_t(spTempElement->getAttribute("RecordTerminator")));
    setTerminator(cValueTemp);
  } //  ENDIF

  //  Tag Info Element <- Get RecordTagInfo Element from File
  //    Description DOM Document
  spTagInfo = spFileDescription->getElementsByTagName(
      "TagInfo")->item[0];

  //  Record ID Field Offset <- Call Tag Info's getAttribute
  //    for Offset
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp,
      _bstr_t(spTagInfo->getAttribute("Offset")) );
  iRecordIDFieldOffset = atoi(cValueTemp);

  //  Record ID Field Length <- Call Tag Info's getAttribute
  //    for Length
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp,
      _bstr_t(spTagInfo->getAttribute("Length")) );
  iRecordIDFieldLength = atoi(cValueTemp);
}


//  Destructor
FlatRecordReader::~FlatRecordReader() {};

//  Get the value of a field
int FlatRecordReader::getFieldValue(int iOffset, int iInLength,
                    char * cField)
{
  //  Locals
  int iLength = iInLength;
  int iStringLen;

  memset(cField,0,iInLength+1);

  //  IF Field Offset > Record Buffer Length
  if (iOffset > iRecordBufferLength)
  {
    //  Return error
    cerr << NO_FIELD_FIXED
         << iOffset << " " << iInLength << endl
         << RecordBuffer << endl;
    return FAILURE;
  } //  ENDIF

  //  IF Field Offset + Field Length > Record Buffer Length
  if ( (iOffset + iInLength) > iRecordBufferLength)
  {
    //  Field Length <- Record Buffer Length - Field Offset
    iLength = iRecordBufferLength - iOffset;
  } //  ENDIF

  //  Field Value <- Extract from Record Buffer according to
  //    Passed Field Offset and adjusted Field Length
  strncpy(cField,&(RecordBuffer[iOffset]),iLength);

  //  Field Value <- Trim trailing white space (<= ASCII space
  //    character)
  iStringLen = strlen(cField);
  while ( (iStringLen > 0) &&
    (cField[iStringLen - 1] <= ' ') )
  {
    cField[iStringLen - 1] = 0;
    iStringLen--;
  }

  //  IF length of Field Value = 0
  if (iStringLen == 0)
  {
    //  Return error
    cerr << NO_FIELD_FIXED
         << iOffset << " " << iInLength << endl
         << RecordBuffer << endl;
    return FAILURE;
  } //  ENDIF

  //  Return Field Value
  return SUCCESS;
}

//  Get the record type
int FlatRecordReader::getRecordType(char * const cRecordType)
{
  return getFieldValue(iRecordIDFieldOffset,
    iRecordIDFieldLength, cRecordType);
}

//  Parse the fields from the input record
int FlatRecordReader::parseRecord(
    const IXMLDOMElementPtr spRecordGrammar)
{
  //  Locals
  //  XML
  IXMLDOMNodeListPtr spFieldGrammarsList;
  IXMLDOMElementPtr spFieldGrammar;

  //  Project
  DataCell * NewCell;

  //  Basic
  char cValueTemp[8];
  int iFieldGrammars, iFieldNumber, iFieldOffset,
    iFieldLengthGrammar, iFieldLength, i;
  char cField[CELL_BUFFER_SIZE];
  int iReturn;

  //  Field Grammar NodeList <- call RecordGrammar's
  //    getElementsByTagName for "FieldDescription"
  spFieldGrammarsList =
    spRecordGrammar->getElementsByTagName("FieldDescription");

  iFieldGrammars = spFieldGrammarsList->length;

  //  DO until end of Field Grammar Node List
  for (i = 0; i < iFieldGrammars; i++)
  {
    //  Field Grammar Element <- Next item in Field Grammar
    //    NodeList
    spFieldGrammar = spFieldGrammarsList->item[i];

    //  Field Number <- call Field Grammar Element's
    //    getAttribute on "FieldNumber"
    memset(cValueTemp,0,sizeof(cValueTemp));
    strcpy(cValueTemp,
        _bstr_t(spFieldGrammar->getAttribute("FieldNumber")) );
    iFieldNumber = atoi(cValueTemp);

    //  Field Offset <- call Field Grammar Element's
    //    getAttribute on "Offset"
    memset(cValueTemp,0,sizeof(cValueTemp));
    strcpy(cValueTemp,
        _bstr_t(spFieldGrammar->getAttribute("Offset")) );
    iFieldOffset = atoi(cValueTemp);

    //  Field Length <- call Field Grammar Element's
    //    getAttribute on "Length"
    memset(cValueTemp,0,sizeof(cValueTemp));
    strcpy(cValueTemp,
        _bstr_t(spFieldGrammar->getAttribute("Length")) );
    iFieldLengthGrammar = atoi(cValueTemp);

    //  IF Field Offset + 1 > Record Buffer Length
    if ( (iFieldOffset + 1) > iRecordBufferLength)
    {
      //  Return success - End of record
      return SUCCESS;
    } //  ENDIF

    iFieldLength = iFieldLengthGrammar;
    //  IF Field Offset + Field Length > Record Buffer Length
    if ( (iFieldOffset + iFieldLengthGrammar) >
          iRecordBufferLength)
    {
      //  Field Length <- Record Buffer Length - Field Offset
      iFieldLength = iRecordBufferLength - iFieldOffset;
    } //  ENDIF

    //  Field Value <- Extract from Record Buffer according to
    //    Field Offset and adjusted Field Length
    memset(cField,0,sizeof(cField));
    memcpy(cField,&(RecordBuffer[iFieldOffset]),iFieldLength);

    //  New Cell <- call RecordHandler's createDataCell method,
    //    passing Field Number and Field Grammar Node
    iReturn = createDataCell(iFieldNumber,spFieldGrammar, NewCell);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Call NewCell's putFieldValue method, passing Field
    //    Value
    iReturn = NewCell->putField(cField,iFieldLength);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
  }
  return SUCCESS;
}

int FlatRecordReader::readRecord()
{
  //  IF Fixed Record Length > 0
  if (iFixedRecordLength > 0)
  {
    //  Return call to ReadRecordFixedLength
    return readRecordFixedLength();
  } //  ENDIF

  //  Return call to base RecordReader's
  //    readRecordVariableLength
  return readRecordVariableLength();
}

int FlatRecordReader::readRecordFixedLength()
{
  //  Clear the record buffer
  memset(RecordBuffer,0,sizeof(RecordBuffer));

  InputFile->read(RecordBuffer, iFixedRecordLength);
	if (InputFile->bad())
  {
    return IO_READ_ERROR;
  }
	if (InputFile->eof())
  {
    return IO_READ_EOF;
  }

  iRecordBufferLength = InputFile->gcount();

  return iRecordBufferLength;
}

