/* FlatRecordWriter.cpp

Handles creating flat file records from XML input

From Chapter 8 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Special headers for this module only
#include "FlatRecordWriter.h"

//  Constructor with output stream
FlatRecordWriter::FlatRecordWriter(
  const IXMLDOMDocument2Ptr spFileDescription,
  ofstream * const OutputFileIn) :
  RecordWriter(spFileDescription, OutputFileIn)
{
  //  Locals
  IXMLDOMElementPtr spRecordFormat;
  IXMLDOMElementPtr spTagInfo;
  IXMLDOMElementPtr spTempElement;
  IXMLDOMNodePtr spTempNode;
  IXMLDOMNodeListPtr spTempNodeList;
  char cValueTemp[3];

  //  Record Format Element <- Get RecordFormat Element from
  //    File Description DOM Document
  spRecordFormat = spFileDescription->getElementsByTagName(
      "RecordFormat")->item[0];

  //  Child Element <- Get first childNode from Record Format Element,
  //    advancing over non-Element Nodes
  spTempNode = spRecordFormat->firstChild;
  while (spTempNode->nodeType != NODE_ELEMENT)
  {
    spTempNode = spTempNode->nextSibling;
  }
  spTempElement = spTempNode;

  //  Fixed Record Length <- 0
  iFixedRecordLength = 0;

  memset(cValueTemp,0,sizeof(cValueTemp));
  //  IF Child Element NodeName = Fixed
  if (!strcmp(spTempElement->nodeName,"Fixed"))
  {
    //  Fixed Record Length <- Call Child Element's
    //    getAttribute for Length
    strcpy(cValueTemp,
        _bstr_t(spTempElement->getAttribute("Length")) );
    iFixedRecordLength = atoi(cValueTemp);
  }
  else  //  ELSE
  {
    //  Record Terminator <- Call Child Element's getAttribute
    //    for RecordTerminator
    strcpy(cValueTemp,
    _bstr_t(spTempElement->getAttribute("RecordTerminator")));
    setTerminator(cValueTemp);
  } //  ENDIF

  //  Tag Info Element <- Get RecordTagInfo Element from File
  //    Description DOM Document
  spTagInfo = spFileDescription->getElementsByTagName(
      "TagInfo")->item[0];

  //  Record ID Field Offset <- Call Tag Info's getAttribute
  //    for Offset
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp,
      _bstr_t(spTagInfo->getAttribute("Offset")) );
  iRecordIDFieldOffset = atoi(cValueTemp);

  //  Record ID Field Length <- Call Tag Info's getAttribute
  //    for Length
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp,
      _bstr_t(spTagInfo->getAttribute("Length")) );
  iRecordIDFieldLength = atoi(cValueTemp);
}

//  Destructor
FlatRecordWriter::~FlatRecordWriter() {};

//  Write record to flat file
int FlatRecordWriter::writeRecord(IXMLDOMElementPtr spRecordGrammar)
{
  //  Locals
  IXMLDOMNodeListPtr spFieldGrammars;
  IXMLDOMElementPtr spFieldGrammar;

  char cGrammarFieldNumber[3];
  char cOffset[8];
  char cLength[8];
  char cFill[2];
  char cRecordTag[CELL_BUFFER_SIZE];

  int iGrammarFieldNumber, iOffset, iLength;
  int iFields, iFieldNumber, iWorkingLength;
  int iGrammarIndex, iFieldIndex;
  char cFieldContents[CELL_BUFFER_SIZE];
  bool boFieldWritten;
  int iReturn;

  memset(cFill,0,sizeof(cFill));
  //  IF Fixed Record Length > 0
  if (iFixedRecordLength > 0)
  {
    //  Initialize Output Record Buffer to ASCII spaces
    cFill[0] = ' ';
    iRecordBufferLength = iFixedRecordLength;
  }
  else
  {
    //  Initialize Output Record Buffer to null characters
    cFill[0] =  0;
  }
  memset(RecordBuffer,cFill[0],iRecordBufferLength);

  //  Record Length <- 0
  iRecordBufferLength = 0;

  //  Field Grammar NodeList <- Call Record Grammar's
  //    GetElementsByTagName for "FieldDescription"
  spFieldGrammars =
    spRecordGrammar->getElementsByTagName("FieldDescription");

  iFields = spFieldGrammars->length;

  //  Cell Array Index <- 0
  //  DO for all items in Field Grammar NodeList
  iFieldIndex = 0;
  for (iGrammarIndex = 0; iGrammarIndex < iFields;
      iGrammarIndex++)
  {
    //  Field Grammar Element <- next item from Field Grammar
    //    NodeList
    spFieldGrammar = spFieldGrammars->item[iGrammarIndex];

    //  Grammar Field Number <- call Field Grammar Element's
    //    GetAttribute for "FieldNumber"
    memset(cGrammarFieldNumber,0,sizeof(cGrammarFieldNumber));
    strcpy(cGrammarFieldNumber,
      _bstr_t(spFieldGrammar->getAttribute("FieldNumber")));
    iGrammarFieldNumber = atoi(cGrammarFieldNumber);

    //  Offset <- Call Grammar Element's getAttribute for
    //    "Offset"
    memset(cOffset,0,sizeof(cOffset));
    strcpy(cOffset,
      _bstr_t(spFieldGrammar->getAttribute("Offset")));
    iOffset = atoi(cOffset);

    //  Length <- Call Grammar Element's getAttribute for
    //    "Length"
    memset(cLength,0,sizeof(cLength));
    strcpy(cLength,
      _bstr_t(spFieldGrammar->getAttribute("Length")));
    iLength = atoi(cLength);

    //  Field Written <- false
    boFieldWritten = false;

    //  IF (Field Index <= Highest Cell )
    if (iFieldIndex <= iHighestCell)
    {
      //  Field Number <- call CellArray[Index] getFieldNumber
      iFieldNumber = CellArray[iFieldIndex]->getFieldNumber();

      //  IF (Field Number = Grammar Field Number)
      if (iFieldNumber == iGrammarFieldNumber)
      {
        //  Call CellArray[Index] fromXML
        iReturn = CellArray[iFieldIndex]->fromXML();
        if (iReturn != SUCCESS)
        {
          return iReturn;
        }

        //  Call CellArray[Index] prepareOutput
        iReturn = CellArray[iFieldIndex]->prepareOutput();
        if (iReturn != SUCCESS)
        {
          return iReturn;
        }

        //  Field Contents <- Call CellArray[Index] getField
        memset(cFieldContents,0,sizeof(cFieldContents));
        CellArray[iFieldIndex]->getField(cFieldContents);

        // Output Buffer <- Insert Field Contents at Offset for
        //    Length
        memcpy(&(RecordBuffer[iOffset]), cFieldContents, iLength);

        //  Clear CellArray[Index]
        delete CellArray[iFieldIndex];

        //  Increment Cell Array Index
        //  Field Written <- true
        boFieldWritten = true;
        iFieldIndex++;
      } //  ENDIF
    } //  ENDIF

    //  IF ( Field Written = false)
    if (!boFieldWritten)
    {
      //  Fill Character <- Call Grammar Element's getAttribute
      //    for "FillCharacter"
      strcpy(cFill,
        _bstr_t(spFieldGrammar->getAttribute("FillCharacter")));

      //  Output Buffer <- Insert Fill Character at Offset for
      //    Length
      memset(&(RecordBuffer[iOffset]), cFill[0], iLength);
    } //  ENDIF

    //  Working Length <- Offset + Length
    iWorkingLength = iOffset + iLength;
    //  IF (Working Length > Record Length)
    if (iWorkingLength > iRecordBufferLength)
    {
      //  Record Length = Working Length
      iRecordBufferLength = iWorkingLength;
    } //  ENDIF
  } //  ENDDO

  //  Highest Cell <- -1
  iHighestCell = -1;

  //  Record ID Value <- Call Record Grammar Element's
  //    getAttribute for "TagValue"
  memset(cRecordTag,0,sizeof(cRecordTag));
  strcpy(cRecordTag,
    _bstr_t(spRecordGrammar->getAttribute("TagValue")));

  //  Output Buffer <- Insert Record ID Value at Record ID
  //    Field Offset for Record ID Field Length
  memcpy(&(RecordBuffer[iRecordIDFieldOffset]), cRecordTag, 
    iRecordIDFieldLength);

  //  IF Fixed Record Length = 0
  if (iFixedRecordLength == 0)
  {
    //  Append base RecordHandler's Record Terminators to
    //    Output Buffer
    RecordBuffer[iRecordBufferLength] = cRecordTerminator1;
    //  Increment Record Buffer Length
    iRecordBufferLength++;
    if ( cRecordTerminator2 != 0 )
    {
      RecordBuffer[iRecordBufferLength] = cRecordTerminator2;
      iRecordBufferLength++;
    }
  }
  else
  {
    iRecordBufferLength = iFixedRecordLength;
  }  //  ENDIF

  //  Call language's write routines to do physical write of
  //    Output Record Buffer
  OutputFile->write(RecordBuffer,iRecordBufferLength);

  if (OutputFile->good())
  {
    return SUCCESS;
  }
  else
  {
    return FAILURE;
  }
}
