/*  FlatSourceConverter.cpp

Drives conversion from flat files to XML

From Chapter 8 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include <direct.h>
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"


//  Program specific header files
#include "FlatRecordReader.h"
#include "FlatSourceConverter.h"
#include <strstrea.h>


//  Default Constructor
FlatSourceConverter::FlatSourceConverter(bool boValidateIn, 
    const char * const cBaseOutputDirectory,
    const char * const cFileDescriptionDocumentName)
    : SourceConverter(boValidateIn, cBaseOutputDirectory)
{
#include "BlasterExceptions.h"
  //  Locals
  IXMLDOMElementPtr spEleTemp;
  IXMLDOMNodeListPtr spNLTemp;
  char cValueTemp[16];

  int i, iReturn;

  //  Initialize Partner Array
  for (i=0; i < MAX_PARTNERS; i++)
  {
    memset(Partners[i],0,sizeof(Partners[i]));
  }
  iNumPartners = 0;

  //  Call loadFileDescriptionDocument from passed document
  iReturn = 
    loadFileDescriptionDocument(cFileDescriptionDocumentName);
  if (iReturn == FAILURE)
  {
    throw cConstructorError;
  }

  //  Schema Location URL <- Call File Description Document's
  //    getElementByTagName on "SchemaLocationURL", then
  //    getAttribute on "value"
  //  NOTE - This is an optional Element, so we may not
  //  get anything
  memset(cSchemaLocationURL,0,sizeof(cSchemaLocationURL));
  spNLTemp = spFileDescription->getElementsByTagName(
      "SchemaLocationURL");
  if (spNLTemp->length != 0)
  {
    spEleTemp = spNLTemp->item[0];
    strcpy(cSchemaLocationURL,
      _bstr_t(spEleTemp->getAttribute("value")));
  }

  //  Throw exception if no schema and validation requested
  if ( boValidate && (strlen(cSchemaLocationURL) == 0) )
  {
    cerr << VALIDATION_WITHOUT_SCHEMA << endl;
    throw cConstructorError;
  }

  //  PartnerBreakOffset <- Get PartnerBreakFieldOffset value
  //    from File Description DOM Document
  //    NOTE:  This is an optional Element so we need to
  //    handle the cases where it might not be present
  //  Partner Break Field Offset <- 0
  //  Partner Break Field Length <- 0
  iPartnerBreakFieldOffset = 0;
  iPartnerBreakFieldLength = 0;

  //  NodeList Temp <- call File Description Document's
  //    getElementsByTagName for "PartnerBreak"
  spNLTemp = spFileDescription->getElementsByTagName(
    "PartnerBreak");

  //  IF (Temp length = 1)
  if (spNLTemp->length == 1)
  {
    //  Partner Break Element <- Temp NodeList item(0)
    spEleTemp = spNLTemp->item[0];
    //  PartnerBreakOffset <- call Partner Break Element's
    //    getAttribute for "Offset", and convert to integer
    memset(cValueTemp,0,sizeof(cValueTemp));
    strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("Offset")));
    iPartnerBreakFieldOffset = atoi(cValueTemp);

    //  PartnerBreakLength <- call Partner Break Element's
    //    getAttribute for "Length", and convert to integer
    memset(cValueTemp,0,sizeof(cValueTemp));
    strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("Length")));
    iPartnerBreakFieldLength = atoi(cValueTemp);
  } //  ENDIF

  //  Initialize Saved Partner
  memset(cSavedPartnerID,0,sizeof(cSavedPartnerID));

  //  Create FlatRecordReader object, passing:
  //    File Description Document DOM
  FlatReader = new FlatRecordReader(spFileDescription);
  Reader = FlatReader;
}


//  Destructor
FlatSourceConverter::~FlatSourceConverter() {};


//  Process a single file
int FlatSourceConverter::processFile(
    const char * const cInputFileName)
{
  //  Locals to method

  //  COM
  HRESULT hResult;
  
  //  XML
  IXMLDOMElementPtr spRecordGrammar;
  IXMLDOMProcessingInstructionPtr spProlog;
  IXMLDOMAttributePtr spSchemaLocationAttribute;

  //  File input and output streams
  ifstream InputFile;

  //  Basic types
  char cOutputDirectory[MAX_PATH_LENGTH];
  char cOutputFilePath[MAX_PATH_LENGTH];
  char cSequenceNumber[4];
  char cSequenceTemp[4];
  char cHeaderTag[CELL_BUFFER_SIZE];
  char cRecordTag[CELL_BUFFER_SIZE];
  int iSequenceNumber, iRecordLength, i, iReturn;
  bool boPartnerFound, boPartnerBreak;

  //  Initializations
  //  Output Directory Path <- Base Directory + directory
  //    separator
  memset(cOutputDirectory,0,sizeof(cOutputDirectory));
  strcpy(cOutputDirectory,cBaseOutputDirectory);
  memset(cOutputFilePath,0,sizeof(cOutputFilePath));
  memset(cHeaderTag,0,sizeof(cHeaderTag));
  spDocOutput = NULL;

  //Initialize Sequence Number
  iSequenceNumber = 0;

  //  Grammar Header Record Tag <- Get Grammar Element's
  //    TagValue Attribute
  strcpy(cHeaderTag, _bstr_t(spGrammar->getAttribute("TagValue")));

  //  Row Grammar <- Get RowDescription from Grammar Element
  spRecordGrammar =
    spGrammar->getElementsByTagName("RowDescription")->item[0];

  //  Open input file
  InputFile.open(cInputFileName, ios::in | ios::nocreate);
  if ( ! InputFile )
  {
    cerr << INPUT_FILE_OPEN_ERROR << cInputFileName << endl;
    return FAILURE;
  }

    //  C++ Only - set mode to binary
  InputFile.setmode(filebuf::binary);


  //  Call CSVRecordReader's setInputStream method
  FlatReader->setInputStream(&InputFile);

  //  Display processing message
  cout << CREATING_DOCUMENT << endl;
  
  //  Call CSVRecordReader's readRecord method
  iRecordLength = FlatReader->readRecord();

  //  Header Tag <- Call FlatRecordReader's getRecordID method
  iReturn = FlatReader->getRecordType(cRecordTag);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  IF (Header Tag != Grammar Header Record Tag)
  if (strcmp(cRecordTag,cHeaderTag))
  {
    //  Return error or throw exception
    cerr << cRecordTag << NOT_HEADER_TAG << cHeaderTag << endl;
    return FAILURE;
  } //  ENDIF

  //  DO until end of input file
  while (iRecordLength >= 0)
  {
    //  PartnerBreak <- Call testPartnerBreak
    //  IF PartnerBreak = true
    iReturn = testPartnerBreak(boPartnerBreak);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
    if ( boPartnerBreak)
    {
      //  Output Directory Path <- Base Directory +
      //    Partner ID + directory separator
      memset(cOutputDirectory,0,sizeof(cOutputDirectory));
      strcpy(cOutputDirectory,cBaseOutputDirectory);
      strcat(cOutputDirectory, cSavedPartnerID);
      strcat(cOutputDirectory, "\\");

      //  Lookup Partner in Partner Array
      boPartnerFound = false;
      for(i = 0; i < iNumPartners; i++)
      {
        if (!strcmp(Partners[i],cSavedPartnerID))
        {
          boPartnerFound = true;
          break;
        }
      }

      //  IF Partner is not in Array
      if (!boPartnerFound)
      {
        //  Partner Array <- Add Partner ID
        strcpy(Partners[iNumPartners],cSavedPartnerID);
        iNumPartners++;

        //  Create output directory from Output Directory
        //  Path
        iReturn = _mkdir(cOutputDirectory);
        if (iReturn != SUCCESS)
        {
          cerr << IO_DIRECTORY_CREATE_ERROR_MESSAGE
               << cOutputDirectory << endl;
          return FAILURE;
        }
      } //  ENDIF

    } //  ENDIF

    //  Create new Output DOM Document
    hResult = spDocOutput.CreateInstance(
      __uuidof(DOMDocument40));

    if FAILED(hResult)
    {
      cerr << CREATE_INSTANCE_ERROR << endl;
      return FAILURE;
    }

    //  Create the header processing instruction
    //    and append it to the document node
    spProlog = spDocOutput->createProcessingInstruction("xml",
      XML_PROLOG_TEXT);
    spDocOutput->appendChild(spProlog);

    //  Error if greater than maximum
    if ((iSequenceNumber+1) > MAX_SEQUENCE_NUMBER)
    {
      cerr << MAX_SEQUENCE_EXCEEDED
           << iSequenceNumber + 1 << endl;
      return FAILURE;
    }
        
    //  Increment Sequence Number
    iSequenceNumber++;

    //  Ouput File Path <- Output directory path +
    //  Root Element Name + Sequence Number + ".xml"
    memset(cSequenceNumber,0,sizeof(cSequenceNumber));
    if ( iSequenceNumber <= 9 )
    {
      strcpy(cSequenceNumber,"00");
      _itoa(iSequenceNumber,cSequenceTemp,10);
      strcat(cSequenceNumber,cSequenceTemp);
    }
    else
    {
      if (iSequenceNumber <= 99)
        {
          strcpy(cSequenceNumber,"0");
          _itoa(iSequenceNumber,cSequenceTemp,10);
          strcat(cSequenceNumber,cSequenceTemp);
        }
        else
        {
          _itoa(iSequenceNumber,cSequenceNumber,10);
        }
    }

    //  Output File Path <- Output directory path + Root
    //    Element Name + Sequence Number + ".xml"
    memset(cOutputFilePath,0,sizeof(cOutputFilePath));
    strcpy(cOutputFilePath, cOutputDirectory);
    strcat(cOutputFilePath,cRootElementName);
    strcat(cOutputFilePath,cSequenceNumber);
    strcat(cOutputFilePath,".xml");

    //  Display document name
    cout << cOutputFilePath << endl;
    
    //  Call FlatRecordReader's setOutputDocument method for
    //    new document
    FlatReader->setOutputDocument(spDocOutput);

    //  Record Length <- Call processGroup to process the
    //    document, passing the Root Element and the Grammar
    //    Element
    iReturn = processGroup(spDocOutput, spGrammar, iRecordLength);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  IF (Record Length > 0)
    if (iRecordLength > 0)
    {
      //  Header Tag <- Call FlatRecordReader's getRecordType
      //    method
      iReturn = FlatReader->getRecordType(cRecordTag);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

      //  IF (Header Tag != Grammar Header Record Tag)
      if (strcmp(cRecordTag,cHeaderTag))
      {
        //  Return error or throw exception
        cerr << cRecordTag << NOT_HEADER_TAG << cHeaderTag << endl;
        return FAILURE;
      } //  ENDIF

    } //  ENDIF

    //  Call saveDocument
    iReturn = saveDocument(spDocOutput,cOutputFilePath,
        boValidate);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    spDocOutput = NULL;

  } //  ENDDO

  // Close input file
  InputFile.close();

  //  Clear Partners Array
  for(i = 0; i < iNumPartners; i++)
  {
    memset(Partners[i],0,sizeof(Partners[i]));
  }

  iNumPartners = 0;

  // Display completion message
  cout << iSequenceNumber << DOCUMENTS_PROCESSED << endl;

  return SUCCESS;
}

//  Test for break on trading partner
int FlatSourceConverter::testPartnerBreak(bool &boPartnerBreak)
{
  //  Local
  int iReturn;
  char cPartnerID[CELL_BUFFER_SIZE];
  memset(cPartnerID,0,sizeof(cPartnerID));

  //  IF Partner Break Legnth is zero
  //    return false
  //  ENDIF
  if (iPartnerBreakFieldLength == 0)
  {
    boPartnerBreak = false;
    return SUCCESS;
  }

  //  PartnerID <- Call FlatRecordReader's getFieldValue
  //    method passing the Partner Break Offset and Partner
  //    Break Length
  iReturn = FlatReader->getFieldValue(iPartnerBreakFieldOffset,
    iPartnerBreakFieldLength, cPartnerID);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  IF DocumentID = Saved Document ID
  if (!strcmp(cPartnerID,cSavedPartnerID))
  {
    //  return false
    boPartnerBreak = false;
    return SUCCESS;
  } //  ENDIF

  //  Saved Partner ID <- Partner ID
  memset(cSavedPartnerID,0,sizeof(cSavedPartnerID));
  strcpy(cSavedPartnerID,cPartnerID);

  //  Return true
  boPartnerBreak = true;
  return SUCCESS;
}

