/* EDIRecordReader.cpp

Base class for reading segments from various EDI syntaxes

From Chapter 9 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Special headers for this module only
#include "EDIRecordReader.h"

//  Default constructor
EDIRecordReader::EDIRecordReader(
    const IXMLDOMDocument2Ptr spFileDescription)
    : RecordReader(spFileDescription)
{
  cElementSeparator = 0;
  cComponentSeparator = 0;
  cRepetitionSeparator = 0;
  cReleaseCharacter = 0;
}


//  Destructor
EDIRecordReader::~EDIRecordReader() {};

//  Get the segment ID or tag from the input segment
int EDIRecordReader::getRecordType(char * const cSegmentTag)
{
  //  Record Tag Value <- Get first characters of input record
  //    up to the element separator
  //  Should not be beyond the fourth character in the buffer
  int iTagLen;
  for (iTagLen=0; iTagLen < 5; iTagLen++)
  {
    if (RecordBuffer[iTagLen] == cElementSeparator)
    {
      break;
    }
  }
  if (iTagLen == 4)
  {
    cerr << NO_SEGMENT_TAG << endl
         << RecordBuffer << endl;
    return FAILURE;
  }

  //  Return Record Tag Value or status
  strncpy(cSegmentTag,RecordBuffer,iTagLen);
  return SUCCESS;
}


//  Parse an EDI segment
int EDIRecordReader::parseRecord(
    const IXMLDOMElementPtr spSegmentGrammar)
{
  //  Locals
  //  XML
  IXMLDOMElementPtr spCurrentGrammar;
  IXMLDOMElementPtr spCompositeGrammar;
  IXMLDOMElementPtr spComponentGrammar;
  IXMLDOMNodePtr spSegmentChild;
  IXMLDOMNodePtr spCompositeChild = NULL;

  //  Locals
  DataCell * CurrentCell = NULL;

  //  Basic
  char cGrammarPosition[4];
  int iGrammarPosition;
  char cGrammarSubPosition[4];
  int iGrammarSubPosition;
  char cNodeName[NMTOKEN_SIZE];
  int iElementPosition;
  int iComponentPosition;
  int iPos;
  char ParsingState;
  char cCurChar;
  int iReturn;

  //  Constants
  const char NextElement = 1;
  const char NextComponent = 2;
  const char NextRepeat = 3;
  const char Escape = 4;
  const char SaveCharacter = 5;
  const char ParsingError = 6;

  //  SegmentChild <- get SegmentGrammar's firstChild
  spSegmentChild = spSegmentGrammar->firstChild;

  //  DO until SegmentChild nodeType is Element
  while (spSegmentChild->nodeType != NODE_ELEMENT)
  {
    //  SegmentChild <- SegmentChild nextSibling
    spSegmentChild = spSegmentChild->nextSibling;
  }
  spCurrentGrammar = spSegmentChild;
  spCompositeGrammar = NULL;

  //  GrammarPosition <- CurrentGrammar getAttribute for
  //    FieldNumber
  memset(cGrammarPosition,0,sizeof(cGrammarPosition));
  strcpy(cGrammarPosition,
    _bstr_t(spCurrentGrammar->getAttribute("FieldNumber")));
  iGrammarPosition = atoi(cGrammarPosition);

  iElementPosition = 0;
  iComponentPosition = 0;
  ParsingState = NextElement;

  //  Advance Pos to point to the RecordBuffer the first
  //     element separator
  iPos = 2;
  while (RecordBuffer[iPos] != cElementSeparator)
  {
    iPos++;
  }

  //  DO while Pos < RecordBufferLength and
  //    ParsingState != ParsingError
  while ( (iPos < iRecordBufferLength) &&
          (ParsingState != ParsingError))
  {
    //  CurChar = RecordBuffer[Pos]
    cCurChar = RecordBuffer[iPos];

    //  Advance to the next state
    //  DO CASE of ParsingState
    switch (ParsingState)
    {
      case NextElement:
        if ( cCurChar == cElementSeparator)
        {
          ParsingState = NextElement;
        }
        else
        {
          if (cCurChar == cComponentSeparator)
          {
            ParsingState = NextComponent;
          }
          else
          {
            if (cCurChar == cRepetitionSeparator)
            {
              ParsingState = NextRepeat;
            }
            else
            {
              if (cCurChar == cReleaseCharacter)
              {
                ParsingState = Escape;
              }
              else
              {
                ParsingState = SaveCharacter;
              }
            }
          }
        }
        break;
      case  NextComponent:
        if (cCurChar == cElementSeparator)
        {
          ParsingState = ParsingError;
        }
        else
        {
          if (cCurChar == cComponentSeparator)
          {
            ParsingState = NextComponent;
          }
          else
          {
            if (cCurChar == cRepetitionSeparator)
            {
              ParsingState = ParsingError;
            }
            else
            {
              if (cCurChar == cReleaseCharacter)
              {
                ParsingState = Escape;
              }
              else
              {
                ParsingState = SaveCharacter;
              }
            }
          }
        }
        break;
      case NextRepeat:
        if (cCurChar == cElementSeparator)
        {
          ParsingState = ParsingError;
        }
        else
        {
          if (cCurChar == cComponentSeparator)
          {
            ParsingState = NextComponent;
          }
          else
          {
            if (cCurChar == cRepetitionSeparator)
            {
              ParsingState = NextRepeat;
            }
            else
            {
              if (cCurChar == cReleaseCharacter)
              {
                ParsingState = Escape;
              }
              else
              {
                ParsingState = SaveCharacter;
              }
            }
          }
        }
        break;
      case Escape:
        ParsingState = SaveCharacter;
        break;
      case SaveCharacter:
        if (cCurChar == cElementSeparator)
        {
          ParsingState = NextElement;
        }
        else
        {
          if (cCurChar == cComponentSeparator)
          {
            ParsingState = NextComponent;
          }
          else
          {
            if (cCurChar == cRepetitionSeparator)
            {
              ParsingState = NextRepeat;
            }
            else
            {
              if (cCurChar == cReleaseCharacter)
              {
                ParsingState = Escape;
              }
              else
              {
                ParsingState = SaveCharacter;
              }
            }
          }
        }
        break;
    }  // END DOCASE

    //  Take action appropriate to state
    //  DO CASE of ParsingState
    switch (ParsingState)
    {
      case NextElement:
        CurrentCell = NULL;
        iElementPosition++;

        // Get the next the grammar Element for the next data
        //   element in the segment
        while (iElementPosition > iGrammarPosition)
        {
          spSegmentChild = spSegmentChild->nextSibling;
          if (spSegmentChild == NULL)
          {
            //  Return error for no grammar element
            cerr << iElementPosition << " "
                 << FIELD_NOT_IN_RECORD_GRAMMAR
                 << _bstr_t(
                     spSegmentGrammar->getAttribute("TagValue"))
                 << endl << RecordBuffer << endl;
            return FAILURE;
          }
          if (spSegmentChild->nodeType != NODE_ELEMENT)
          {
            continue;
          }
          spCurrentGrammar = spSegmentChild;
          memset(cGrammarPosition,0,sizeof(cGrammarPosition));
          strcpy(cGrammarPosition, _bstr_t(
              spCurrentGrammar->getAttribute("FieldNumber")));
          iGrammarPosition = atoi(cGrammarPosition);
        } //  ENDDO
        //  If we find a match then do the setup, otherwise set
        //    the current grammar Element to NULL
        if ( iElementPosition == iGrammarPosition)
        {
          memset(cNodeName,0,sizeof(cNodeName));
          strcpy(cNodeName, spSegmentChild->nodeName);
          if (!strcmp(cNodeName,"CompositeStructureDescription"))
          {
            spCompositeGrammar = spSegmentChild;
            spCompositeChild =
              spCompositeGrammar->firstChild;
            //  DO until CompositeChild nodeType is Element
            while (spCompositeChild->nodeType != NODE_ELEMENT)
            {
              spCompositeChild = spCompositeChild->nextSibling;
            }
            spCurrentGrammar = spCompositeChild;
            iComponentPosition = 1;
          }
          else
          {
            spCompositeGrammar = NULL;
            spCompositeChild = NULL;
            iComponentPosition = 0;
            spCurrentGrammar = spSegmentChild;
          }
        }
        else
        {
          spCurrentGrammar = NULL;
        }
        break;
      case NextComponent:
        CurrentCell = NULL;
        iComponentPosition++;
        //  GrammarSubPosition <- CurrentGrammar getAttribute
        //    for SubFieldNumber
        memset(cGrammarSubPosition,0,sizeof(cGrammarSubPosition));
        strcpy(cGrammarSubPosition, _bstr_t(
            spCurrentGrammar->getAttribute("SubFieldNumber")));
        iGrammarSubPosition = atoi(cGrammarSubPosition);

        // Get the next the grammar Element for the next data
        //   element in the Composite
        //  DO while ComponentPosition > GrammarSubPosition
        while (iComponentPosition > iGrammarSubPosition)
        {
          spCompositeChild = spCompositeChild->nextSibling;
          if (spCompositeChild == NULL)
          {
            //  Return error for no grammar element
            cerr << iElementPosition << " " << iComponentPosition
                 << " " << FIELD_NOT_IN_RECORD_GRAMMAR
                 << _bstr_t(
                     spSegmentGrammar->getAttribute("TagValue"))
                 << endl << RecordBuffer << endl;
            return FAILURE;
          }
          if (spCompositeChild->nodeType != NODE_ELEMENT)
          {
            continue;
          }

          spComponentGrammar = spCompositeChild;
          //  GrammarSubPosition <- CurrentGrammar getAttribute
          //    for SubFieldNumber
          memset(cGrammarSubPosition,0,sizeof(cGrammarSubPosition));
          strcpy(cGrammarSubPosition, _bstr_t(
            spComponentGrammar->getAttribute("SubFieldNumber")));
          iGrammarSubPosition = atoi(cGrammarSubPosition);
        } //  ENDDO
        //  If we find a match then do the setup, otherwise set
        //    the current grammar Element to NULL
        if (iComponentPosition == iGrammarSubPosition)
        {
          spCurrentGrammar = spCompositeChild;
        }
        else
        {
          spCurrentGrammar = NULL;
        }
        break;
      case NextRepeat:
        // If we're processing a composite, reset the grammar and
        //   index back to the start of the composite.  Otherwise
        //   do nothing and stay with the current grammar
        CurrentCell = NULL;
        if (spCompositeGrammar != NULL)
        {
          spCompositeChild = spCompositeGrammar->firstChild;
          while (spCompositeChild->nodeType != NODE_ELEMENT)
          {
            spCompositeChild =
                spCompositeChild->nextSibling;
          }
          spCurrentGrammar = spCompositeChild;
          iComponentPosition = 1;
        }
        break;
      case Escape:
        //  Take no action
        break;
      case SaveCharacter:
        if (CurrentCell == NULL)
        {
          if (spCurrentGrammar == NULL)
          {
            //  Return error for no grammar Element
            cerr << iElementPosition << " " << iComponentPosition
                 << " " << FIELD_NOT_IN_RECORD_GRAMMAR
                 << _bstr_t(
                     spSegmentGrammar->getAttribute("TagValue"))
                 << endl << RecordBuffer << endl;
            return FAILURE;
          }
          //  CurrentCell = call createDataCell, passing
          //    ElementPosition and CurrentGrammar
          iReturn = createDataCell(iElementPosition,
              spCurrentGrammar, CurrentCell);
          if (iReturn != SUCCESS)
          {
            return iReturn;
          }
          if (iComponentPosition != 0)
          {
            //  Call DataCell's setSubFieldNumber, passing
            //    ComponentPosition
            CurrentCell->setSubFieldNumber(iComponentPosition);
          }
        }
        //  Call DataCell putByte, passing CurChar
        iReturn = CurrentCell->putByte(cCurChar);
        if (iReturn != SUCCESS)
        {
          return iReturn;
        }
        break;
      case ParsingError:
        //  Take no action
        break;
    } //  ENDDO CASE
    iPos++;
  } //  ENDDO

  //  Check if we are finishing with a delimiter
  if (ParsingState != SaveCharacter)
  {
    ParsingState = ParsingError;
  }

  if (ParsingState == ParsingError)
  {
    //  Display parsing error message with ElementPosition,
    //    ComponentPosition, Pos, and segment contents
    //  Return error
    cerr << EDI_PARSING_ERROR << endl
         << POSITION_IN_SEGMENT << iElementPosition << endl
         << POSITION_IN_COMPOSITE << iComponentPosition << endl
         << CHARACTER_POSITION_IN_SEGMENT << iPos << endl
         << RecordBuffer << endl;
    return FAILURE;
  }

 //  Return success
  return SUCCESS;
  
}


//  We define this method for use with a base class pointer
int EDIRecordReader::readRecord()
{
  return readRecordVariableLength();
}

//  Set the File Description Document
void EDIRecordReader::setFileDescriptionDocument(
    IXMLDOMDocument2Ptr spFDD)
{
  spFileDescription = spFDD;
}

//  Write a Record to the output XML Document from the DataCell
//  Array - Same logic as base RecordReader class, but handles
//  composite data structures in EDI segments
void EDIRecordReader::writeRecord(IXMLDOMElementPtr spParent,
  IXMLDOMElementPtr spSegmentGrammar)
{
  IXMLDOMElementPtr spRecord;
  IXMLDOMElementPtr spCurrentParent;
  IXMLDOMElementPtr spCompositeParent;
  IXMLDOMElementPtr spCompositeGrammar;
  IXMLDOMElementPtr spField;
  IXMLDOMNodeListPtr spCompositeGrammarsList;
  char cElementName[NMTOKEN_SIZE], cCompositePosition[3];
  int i, j, iCompositePosition,
    iComponentPosition, iCellFieldNumber;

  //  ElementName <- Call Grammar Element's getAttribute for
  //    "ElementName"
  memset(cElementName,0,sizeof(cElementName));
  strcpy(cElementName,
      _bstr_t(spSegmentGrammar->getAttribute("ElementName")));

  //  Record Element <- call Output Document's createElement
  spRecord = spDocOutput->createElement(cElementName);

  //  Parent Element <- call Parent's appendChild to append
  //    Record Element
  spParent->appendChild(spRecord);

  //  CurrentParent <- RecordElement
  spCurrentParent = spRecord;

  //  CompositeParent <- null
  spCompositeParent = NULL;

  //  CompositePosition <- 0
  iCompositePosition = 0;

  //  CompositeGrammars <- call RecordGrammar's
  //   getElementsByTagName for "CompositeStructureDescription"
  spCompositeGrammarsList = spSegmentGrammar->getElementsByTagName(
                              "CompositeStructureDescription");

  //  DO for all DataCells in array up to Highest Cell
  for (i = 0; i <= iHighestCell; i++)
  {
    // Test if this is a component within a composite
    //  ComponentPosition <- call DataCell's getSubFieldNumber
    iComponentPosition = CellArray[i]->getSubFieldNumber();

    //  IF ComponentPosition != 0
    if (iComponentPosition != 0)
    {
      //  See if we are starting a new composite
      //  CellFieldNumber <- Call DataCell's getFieldNumber
      iCellFieldNumber = CellArray[i]->getFieldNumber();
      //  IF CellFieldNumber != CompositePosition
      if (iCellFieldNumber != iCompositePosition)
      {
        // Find the grammar and create the Element
        //  DO for all Elements in CompositeGrammars NodeList
        //    until CellFieldNumber = CompositePosition
        j = -1;
        do
        {
          j++;
          //  CompositePosition <- Call CompositeGrammar item's
          //    GetAttribute for "FieldNumber"
          spCompositeGrammar =
               spCompositeGrammarsList->item[j];
          memset(cCompositePosition,0,
              sizeof(cCompositePosition));
          strcpy(cCompositePosition, _bstr_t(
              spCompositeGrammar->getAttribute("FieldNumber")));
          iCompositePosition = atoi(cCompositePosition);
        } while (iCompositePosition != iCellFieldNumber);

        //  CompositeName <- Call CompositeGrammars current
        //    item Element's getAttribute for "ElementName"
        memset(cElementName,0,sizeof(cElementName));
        strcpy(cElementName, _bstr_t(
            spCompositeGrammar->getAttribute("ElementName")));

        //  CompositeParent <- call Output Document's
        //    createElement, passing CompositeName
        spCompositeParent =
            spDocOutput->createElement(cElementName);
        //  RecordElement <- call RecordElement appendChild to
        //    append CompositeParent
        spRecord->appendChild(spCompositeParent);
        //  CurrentParent <- CompositeParent
        spCurrentParent = spCompositeParent;
      } //  ENDIF
    }
    else  //  ELSE
    {
      //  Simple element
      //  CurrentParent <- RecordElement
      spCurrentParent = spRecord;
      //  CompositeParent <- null
      spCompositeParent = NULL;
      //  CompositePosition <- 0
      iCompositePosition = 0;
    }
    //  Call toElement on Cell Array entry, passing
    //    CurrentParent
    CellArray[i]->toElement(spDocOutput,spCurrentParent);
    //  Clear Cell Array Entry
    delete CellArray[i];
  } //  ENDDO

  //  Clear highest cell back to -1
  iHighestCell = -1;
  return;
}

