/* EDIRecordWriter.cpp

Base class for creating EDI segments from XML input

From Chapter 9 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"

//  Special headers for this module only
#include "EDIRecordWriter.h"

//  Constructor with output stream
EDIRecordWriter::EDIRecordWriter(
  const IXMLDOMDocument2Ptr spFileDescription,
  ofstream * const OutputFileIn) :
  RecordWriter(spFileDescription, OutputFileIn)
{
#include "BlasterExceptions.h"

    //  Locals
  //  COM
  HRESULT hResult;

  // XML
  IXMLDOMElementPtr spDelimiters;
  IXMLDOMElementPtr spEleTemp;
  IXMLDOMNodeListPtr spNodeListTemp;
  IXMLDOMParseErrorPtr spParseError;

  // Basic
  char cValueTemp[3];
  char cBBHomePath[MAX_PATH_LENGTH];
  char * cTemp;
  int iLen;

  //  Delimiters Element <- Call File Description Document's
  //    getElementsByTagName for "Delimiters"
  spDelimiters = spFileDescription->getElementsByTagName(
    "Delimiters")->item[0];

  //  SegmentTerminator Element <- call Delimiters
  //    getElementsByTagName for "SegmentTerminator"
  //  Delimiter Value = call SegmentTerminator's getAttribute
  //    for "value"
  //  Record Terminator1 <- Call setDelimiter, passing
  //    Delimiter Value
  spEleTemp = spDelimiters->getElementsByTagName(
    "SegmentTerminator")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")));

  cRecordTerminator1 = setDelimiter(cValueTemp);

  //  ElementSeparator  <- Get ElementSeparator value from File
  //      Description DOM Document
  spEleTemp = spFileDescription->getElementsByTagName(
    "ElementSeparator")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")));
  cElementSeparator = setDelimiter(cValueTemp);

  //  ComponentSeparator  <- Get ComponentSeparator value from
  //      File Description DOM Document
  spEleTemp = spFileDescription->getElementsByTagName(
    "ComponentSeparator")->item[0];
  memset(cValueTemp,0,sizeof(cValueTemp));
  strcpy(cValueTemp, _bstr_t(spEleTemp->getAttribute("value")));
  cComponentSeparator = setDelimiter(cValueTemp);

  //  RepetitionSeparator Element <- call Delimiters
  //      getElementsByTagName for "RepetitionSeparator"
  //  IF RepetitionSeparator Element is not null
  //    Delimiter Value = call RepetitionSeparator Element's
  //      getAttribute for "value"
  //    RepetitionSeparator <- Call setDelimiter, passing
  //      Delimiter Value
  //  ENDIF
  spNodeListTemp = spFileDescription->getElementsByTagName(
    "RepetitionSeparator");
  if(spNodeListTemp->length != 0)
  {
    spEleTemp = spNodeListTemp->item[0];
    memset(cValueTemp,0,sizeof(cValueTemp));
    strcpy(cValueTemp,
        _bstr_t(spEleTemp->getAttribute("value")));
    cRepetitionSeparator = setDelimiter(cValueTemp);
  }
  else
  {
    cRepetitionSeparator = 0;
  }

  //  ReleaseCharacter Element <- call Delimiters
  //    getElementsByTagName for "ReleaseCharacter"
  //  IF ReleaseCharacter Element is not null
  //    Delimiter Value = call ReleaseCharacter Element's
  //      getAttribute for "value"
  //    ReleaseCharacter <- Call setDelimiter, passing
  //      Delimiter Value
  //  ENDIF
  spNodeListTemp = 
    spFileDescription->getElementsByTagName("ReleaseCharacter");
  if(spNodeListTemp->length != 0)
  {
    spEleTemp = spNodeListTemp->item[0];
    memset(cValueTemp,0,sizeof(cValueTemp));
    strcpy(cValueTemp,
        _bstr_t(spEleTemp->getAttribute("value")));
    cReleaseCharacter = setDelimiter(cValueTemp);
  }
  else
  {
    cReleaseCharacter = 0;
  }

  //  BBHomePath <- from system property or environment
  //    variable for BBHome
  
  cTemp = getenv("BBHOME");
  if (cTemp == NULL )
  {
    cerr << BBHOME_NOT_SET << endl;
    throw cConstructorError;   
  }
  memset(cBBHomePath,0,sizeof(cBBHomePath));
  strcpy(cBBHomePath,cTemp);
  iLen = strlen(cBBHomePath);

  if ( cBBHomePath[iLen - 1] != '\\' )
  {
    cBBHomePath[iLen] = '\\';
  }

  //  EDIControlPath <- BBHomePath + "EDIControl.xml"
  memset(cEDIControlPath,0,sizeof(cEDIControlPath));
  strcpy(cEDIControlPath,cBBHomePath);
  strcat(cEDIControlPath, "EDIControl.xml");

  //  Call implementation dependent methods to load EDIControl
  //    Document
  //  Create the COM DOM Document object
  hResult = spEDIControl.CreateInstance(__uuidof(DOMDocument40));

  if FAILED(hResult)
  {
    throw cCreateInstanceError;
  }

  //  Tell it we don't want to load the document asynchronously
  spEDIControl->async =  VARIANT_FALSE; 

  //  Make sure that we validate on parse
  spEDIControl->validateOnParse = VARIANT_TRUE;

  //  Load it
  hResult = spEDIControl->load(cEDIControlPath);
  
  // Check for errors
  if( hResult != VARIANT_TRUE)
  {
    spParseError = spEDIControl->GetparseError();
    cerr << PARSE_ERROR_BRIEF << cEDIControlPath
        << endl;
    displayParseError(spParseError);
    throw cConstructorError;   
  }
}

//  Destructor
EDIRecordWriter::~EDIRecordWriter() {};


//  Process a record from the input XML document into the
//  DataCell Array - similar to method in RecordWriter
//  base class, but handles composite data structures
int EDIRecordWriter::parseRecord(IXMLDOMElementPtr spRecord,
    IXMLDOMElementPtr spRecordGrammar)
{
  IXMLDOMElementPtr spFieldElement;
  IXMLDOMNodePtr spFieldNode;
  IXMLDOMElementPtr spFieldElementGrammar;
  IXMLDOMNodePtr spFieldNodeGrammar;
  IXMLDOMNodePtr spChild;

  DataCell * NewCell;

  char cFieldName[NMTOKEN_SIZE];
  char cGrammarFieldName[NMTOKEN_SIZE];

  char cFieldNumber[4];
  char cElementText[CELL_BUFFER_SIZE];
  int iFieldNumber;
  bool boComposite;
  int iReturn;

  memset(cElementText,0,sizeof(cElementText));    

  //  Field Grammar <- Get first Grammar child Element from
  //  Record Grammar Element, skipping over non-Element nodes
  spFieldNodeGrammar = spRecordGrammar->firstChild;
  while (spFieldNodeGrammar->nodeType != NODE_ELEMENT)
  {
    spFieldNodeGrammar = spFieldNodeGrammar->nextSibling;
  }
  spFieldElementGrammar = spFieldNodeGrammar;

  //  Grammar Field Name <- call Field Grammar's getAttribute
  //  on "ElementName"
  memset(cGrammarFieldName,0,sizeof(cGrammarFieldName));
  strcpy(cGrammarFieldName,
    _bstr_t(spFieldElementGrammar->getAttribute("ElementName")));

  //  Get the first Field child Element of the data Record
  spFieldNode = spRecord->firstChild;
  //  DO for all Field Elements in data Record's NodeList
  while (spFieldNode != NULL)
  {
    //  Skip any non-Element Nodes
    if (spFieldNode->nodeType != NODE_ELEMENT)
    {
      spFieldNode = spFieldNode->nextSibling;
      continue;
    }

    spFieldElement = spFieldNode;

    // Test for a composite data structure
    //  Composite <- False
    boComposite = false;

    //  childNode <- get Field Element's firstChild
    spChild = spFieldElement->firstChild;

    //  DO until childNode is null or Composite is true
    while ( (spChild != NULL) && (boComposite == false) )
    {
      if (spChild->nodeType == NODE_ELEMENT)
      {
        boComposite = true;
      }
      else
      {
        spChild = spChild->nextSibling;
      }
    } // END DO

    //  IF Composite is false
    if (!boComposite)
    {
      //  ElementText <- call getElement Text on Field Element
      memset(cElementText,0,sizeof(cElementText));    
      iReturn = getElementText(spFieldElement, cElementText,
                sizeof(cElementText));
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

      //  IF Element Text is empty
      if (strlen(cElementText) == 0)
      {
        //  Proceed to next field
        spFieldNode = spFieldNode->nextSibling;
        continue;
      } //  ENDIF
    } //  ENDIF composite is false

    //  Field Name <- Field Element's tagName attribute
    memset(cFieldName,0,sizeof(cFieldName));
    strcpy(cFieldName, spFieldElement->tagName);

    //  DO UNTIL Field Name != Grammar Field Name
    while (strcmp(cGrammarFieldName,cFieldName))
    {
      //  Get next sibling Field Grammar Element
      spFieldNodeGrammar = spFieldNodeGrammar->nextSibling;

      //  Skip over non-Element Nodes
      while (spFieldNodeGrammar->nodeType != NODE_ELEMENT)
      {
        spFieldNodeGrammar = spFieldNodeGrammar->nextSibling;
        if (spFieldNodeGrammar == NULL)
        {
          char cRecordElementName[NMTOKEN_SIZE];
          memset(cRecordElementName,0,
              sizeof(cRecordElementName));
          // Throw exception for Element not in Record Grammar
          strcpy(cRecordElementName,
            _bstr_t(
                spRecordGrammar->getAttribute("ElementName")));
          cerr << cFieldName
               << FIELD_ELEMENT_NOT_IN_RECORD_GRAMMAR
               << cRecordElementName << endl;
          return FAILURE;
        }
      }
      spFieldElementGrammar = spFieldNodeGrammar;

      //  Grammar Field Name <- call Field Grammar's
      //    getAttribute on "ElementName"
      memset(cGrammarFieldName,0,sizeof(cGrammarFieldName));
      strcpy(cGrammarFieldName,
        _bstr_t(
          spFieldElementGrammar->getAttribute("ElementName")));
    }

    //  IF Composite is true
    if (boComposite)
    {
      //  Call base class parseRecord, passing Field Element
      //  and Grammar Element
      RecordWriter::parseRecord(spFieldElement,
          spFieldElementGrammar);
    }
    else
    {
      //  Field Number <- Get call Field Grammar's getAttribute
      //    on "Number"
      memset(cFieldNumber,0,sizeof(cFieldNumber));
      strcpy(cFieldNumber,
          _bstr_t(
            spFieldElementGrammar->getAttribute("FieldNumber")));
      iFieldNumber = atoi(cFieldNumber);

      //  NewCell <- call createDataCell, passing Field Number
      //    and DataType
      iReturn = createDataCell(iFieldNumber,
          spFieldElementGrammar, NewCell);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

      //  NewCell <- call NewCell's putField
      NewCell->putField(cElementText,strlen(cElementText));
    }

    //  Get next field Element from Record Element
    spFieldNode = spFieldNode->nextSibling;
  } //  ENDDO for all child Elements of Record Element

  return SUCCESS;
}

//  Write an EDI Segment from the Data Cell array
int EDIRecordWriter::writeRecord(
                IXMLDOMElementPtr spSegmentGrammar)
{
  //  Locals
  int i, iCellPosition, iCellSubPosition, iElementPosition,
    iComponentPosition, iPreviousElementPosition, iLen;
  int iReturn;
  char cSegmentTag[4];
  char cCellContents[CELL_BUFFER_SIZE];

  memset(RecordBuffer,0,iRecordBufferLength);
  //  Record Length <- 0
  iRecordBufferLength = 0;

  //  SegmentTag <- Call Segment Grammar's getAttribute for
  //    "TagValue"
  memset(cSegmentTag,0,sizeof(cSegmentTag));
  strcpy(cSegmentTag,
      _bstr_t(spSegmentGrammar->getAttribute("TagValue")));
  //  OutputBuffer <- SegmentTag
  strcpy(RecordBuffer,cSegmentTag);
  iLen = strlen(cSegmentTag);

  //  Set Record Buffer Length to Tag length
  iRecordBufferLength = iLen;

  iElementPosition = 0;
  iPreviousElementPosition = 0;
  iComponentPosition = 1;

  //  DO for 0 through Highest Cell
  for (i=0; i <= iHighestCell; i++)
  {
    //  Test for new data element
    //  Cell Column Number = Get Cell's Field Number
    iCellPosition = CellArray[i]->getFieldNumber();
    // Reset to start new composite if needed
    if (iElementPosition != iCellPosition)
    {
      iComponentPosition = 1;
    }
    // Write preceding data element separators
    //  DO while Output Column Number < Cell Column Number
    while (iElementPosition < iCellPosition)
    {
      //  Append ElementSeparator to Output Record Buffer
      RecordBuffer[iRecordBufferLength] = cElementSeparator;
      iRecordBufferLength++;
      RecordBuffer[iRecordBufferLength] = 0;
      //  Increment ElementPosition
      iElementPosition++;
    } //  ENDIF

    // Write preceding repetition separator
    // NOTE - An additional test is needed here to write
    //    composite data structures correctly
    //    For now we'll just skip
    if ( (iElementPosition == iPreviousElementPosition) && 0)
    {
      //  Append RepetitionSeparator to Output Record Buffer
      RecordBuffer[iRecordBufferLength] = cRepetitionSeparator;
      iRecordBufferLength++;
    }
    iPreviousElementPosition = iElementPosition;

    // Write preceding component data element separators
    iCellSubPosition = CellArray[i]->getSubFieldNumber();
    while (iComponentPosition < iCellSubPosition)
    {
      //  Append ComponentSeparator to Output Record Buffer
      RecordBuffer[iRecordBufferLength] = cComponentSeparator;
      iRecordBufferLength++;
      iComponentPosition++;
    }
    // Call fromXML to convert Cell contents from XML datatype
    iReturn = CellArray[i]->fromXML();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Call prepareOutput to format Cell contents for EDI
    //    output
    iReturn = CellArray[i]->prepareOutput();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Cell Contents <- Call getField to retrieve cell buffer
    //    contents
    memset(cCellContents,0,sizeof(cCellContents));
    iLen = CellArray[i]->getField(cCellContents);

    //  Clear Array entry
    delete CellArray[i];

    //  Append Cell Contents to Output Record Buffer
    strcat(RecordBuffer,cCellContents);
    iRecordBufferLength = iRecordBufferLength + iLen;
  }   // ENDDO

  iHighestCell = -1;

  //  Append base Segment Terminator to Output
  //    Record Buffer
  RecordBuffer[iRecordBufferLength] = cRecordTerminator1;
  iRecordBufferLength++;

  //  Call language's write routines to do physical write of
  //    Output Record Buffer
  OutputFile->write(RecordBuffer,iRecordBufferLength);

  if (OutputFile->good())
  {
    iSegmentCount++;
    return SUCCESS;
  }
  else
  {
    return FAILURE;
  }
}

int EDIRecordWriter::writeEDIControl()
{
  //  COM
  HRESULT hResult;

  hResult = spEDIControl->save(cEDIControlPath);
  
  // Check for errors
  if FAILED(hResult)
  {
    cerr << XML_OUTPUT_SAVE_ERROR << cEDIControlPath << endl;
    return FAILURE;
  }  
  return SUCCESS;
}
