/* X12RecordReader.cpp

Class for reading control segments from X12 interchanges

From Chapter 9 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Special headers for this module only
#include "EDIRecordReader.h"
#include "X12RecordReader.h"

//  Default constructor
X12RecordReader::X12RecordReader(
    const IXMLDOMDocument2Ptr spFileDescription)
    : EDIRecordReader(spFileDescription)
{
  //  Index in each of these arrays corresponds to the
  //  element postion within the segment without 0 offset.
  //  For example, GS01 is in GS[1]
  int i;
  for(i=0; i < 17; i++)
  {
    memset(ISA[i],0,sizeof(ISA[i]));
  }
  for(i=0; i < 3; i++)
  {
    memset(IEA[i],0,sizeof(IEA[i]));
  }
  for(i=0; i < 9; i++)
  {
    memset(GS[i],0,sizeof(GS[i]));
  }
  for(i=0; i < 3; i++)
  {
    memset(GE[i],0,sizeof(GE[i]));
  }
  for(i=0; i < 4; i++)
  {
    memset(ST[i],0,sizeof(ST[i]));
  }
  for(i=0; i < 3; i++)
  {
    memset(SE[i],0,sizeof(SE[i]));
  }
}

//  Destructor
X12RecordReader::~X12RecordReader() {};

//  Get the segment ID or tag from the input segment
int X12RecordReader::getControlSegmentElement(
    const char * const cTag, int iPosition, char * const cContent)
{
  bool boBad = true;
  if (!(strcmp(cTag,"ISA")))
  {
    strcpy(cContent,ISA[iPosition]);
    boBad = false;
  }
  if (!(strcmp(cTag,"GS")))
  {
    strcpy(cContent,GS[iPosition]);
    boBad = false;
  }
  if (!(strcmp(cTag,"ST")))
  {
    strcpy(cContent,ST[iPosition]);
    boBad = false;
  }
  if (!(strcmp(cTag,"GE")))
  {
    strcpy(cContent,GE[iPosition]);
    boBad = false;
  }

  if (boBad)
  {
    cerr << "Programming error, you dummy!" << endl
         << "You asked for " << cTag << " " << iPosition << endl;
    return FAILURE;
  }
  return SUCCESS;
}


//  Methods for X12 Interchange control segments
//  Read the interchange control header segment from the
//  input file and parse
int X12RecordReader::parseISA()
{
  int i;
  char ISASegment[128];
  char cTemp[128];
  char * cToken;
  // RecordBuffer <- call Lanuage specific routines to read 106
  //   bytes from input file stream
  // RecordBufferLength <- 106
  memset(ISASegment,0,sizeof(ISASegment));
  memset(cElementSeparatorString,0,
      sizeof(cElementSeparatorString));

  InputFile->read(ISASegment, 106);
  if (InputFile->bad())
  {
    return IO_READ_ERROR;
  }
  if (InputFile->eof())
  {
    return IO_READ_EOF;
  }

  iRecordBufferLength = InputFile->gcount();

  if (iRecordBufferLength != 106)
  {
    cerr << NOT_INTERCHANGE_HEADER << "X12" << endl
         << ISASegment << endl;
    return FAILURE; 
  }

  //  IF character at RecordBuffer index 93 = '-'
  //  Get one more byte from input file and append to Record
  //  Buffer - minus sign is not included in min/max length
  //  of control number
  //  RecordBufferLength <- 107
  if (RecordBuffer[93] == '-')
  {
    InputFile->get(ISASegment[106]);
    if (InputFile->bad())
    {
      return IO_READ_ERROR;
    }
    iRecordBufferLength++; 
  }

  //  If 107th byte not read or
  //  IF RecordBuffer doesn't start with "ISA"
  //  Return Error for not an ISA segment
  if (strncmp(ISASegment,"ISA",3))
  {
    cerr << NOT_INTERCHANGE_HEADER << "X12" << endl
         << ISASegment << endl;
    return FAILURE; 
  }

  cElementSeparator = ISASegment[3];
  cComponentSeparator = ISASegment[iRecordBufferLength - 2];
  cRecordTerminator1 = ISASegment[iRecordBufferLength - 1];
  cRecordTerminator2 = 0;

  //  Using language specific string token routines,
  //  load class attributes ISA01 through ISA16
  //  For C++ we copy to temporary string since strtok modifies
  //  the parsed string
  memset(cTemp,0,sizeof(cTemp));
  cElementSeparatorString[0] = cElementSeparator;
  strcpy(cTemp,ISASegment);

  cToken = strtok(cTemp,cElementSeparatorString);
  for (i=1; i < 17; i++)
  {
    cToken = strtok(NULL,cElementSeparatorString);
    if (cToken == NULL)
    {
      cerr << INSUFFICIENT_ELEMENTS << endl
           << "Expected:  16  Found: " << i << " in ISA" << endl
           << ISASegment << endl;
      return FAILURE;
    }
    strcpy(ISA[i],cToken);
  }
  return SUCCESS;
}

int X12RecordReader::parseIEA()
{
  int i;
  char cTemp[32];
  char * cToken;
  memset(cTemp,0,sizeof(cTemp));
  strcpy(cTemp,RecordBuffer);

  cToken = strtok(cTemp,cElementSeparatorString);
  for (i=1; i < 3; i++)
  {
    cToken = strtok(NULL,cElementSeparatorString);
    if (cToken == NULL)
    {
      cerr << INSUFFICIENT_ELEMENTS << endl
           << "Expected:  2  Found: " << i << " in IEA"<< endl
           << RecordBuffer << endl;
      return FAILURE;
    }
    strcpy(IEA[i],cToken);
  }
  return SUCCESS;
}

int X12RecordReader::logISA()
{
  cout << "** Processing X12 Interchange ** " << endl
    << "    Sender ID:         " <<  ISA[6] << endl
    << "    Receiver ID:       " <<  ISA[8] << endl
    << "    Date:              " <<  ISA[9] << endl
    << "    Time:              " << ISA[10] << endl
    << "    Control Number:    " <<  ISA[13] << endl
    << "    Production/Test:   " <<  ISA[15] << endl << endl;
  return SUCCESS;
}

int X12RecordReader::logIEA()
{
  cout << "Processed X12 Interchange" << endl
    << "    Included Groups    " <<  IEA[1] << endl
    << "    Control Number:    " <<  IEA[2] << endl
    << endl;
  return SUCCESS;
}

int X12RecordReader::parseGS()
{
  int i;
  char cTemp[128];
  char * cToken;
  memset(cTemp,0,sizeof(cTemp));
  strcpy(cTemp,RecordBuffer);
  cToken = strtok(cTemp,cElementSeparatorString);

  for (i=1; i < 9; i++)
  {
    cToken = strtok(NULL,cElementSeparatorString);
    if (cToken == NULL)
    {
      cerr << INSUFFICIENT_ELEMENTS << endl
           << "Expected:  8  Found: " << i << " in GS"<< endl
           << RecordBuffer << endl;
      return FAILURE;
    }
    strcpy(GS[i],cToken);
  }
  return SUCCESS;
}

int X12RecordReader::parseGE()
{
  int i;
  char cTemp[32];
  char * cToken;
  memset(cTemp,0,sizeof(cTemp));
  strcpy(cTemp,RecordBuffer);

  cToken = strtok(cTemp,cElementSeparatorString);
  for (i=1; i < 3; i++)
  {
    cToken = strtok(NULL,cElementSeparatorString);
    if (cToken == NULL)
    {
      cerr << INSUFFICIENT_ELEMENTS << endl
           << "Expected:  2  Found: " << i << " in GE"<< endl
           << RecordBuffer << endl;
      return FAILURE;
    }
    strcpy(GE[i],cToken);
  }
  return SUCCESS;
}

int X12RecordReader::logGS()
{
  cout << "  Processing Functional Group\n"
    << "      Group Type ID:     " <<  GS[1] << endl
    << "      Sender ID:         " <<  GS[2] << endl
    << "      Receiver ID:       " <<  GS[3] << endl
    << "      Date:              " <<  GS[4] << endl
    << "      Time:              " <<  GS[5] << endl
    << "      Control Number:    " <<  GS[6] << endl
    << "      Agency:            " <<  GS[7] << endl
    << "      Version/Release:   " <<  GS[8] << endl
    << endl;
  return SUCCESS;
}

int X12RecordReader::logGE()
{
  cout << "  Processed Functional Group\n"
    << "      Included Sets:     " <<  GE[1] << endl
    << "      Control Number:    " <<  GE[2] << endl
    << endl;
  return SUCCESS;
}

int X12RecordReader::parseST()
{
  int i;
  char cTemp[64];
  char * cToken;
  memset(cTemp,0,sizeof(cTemp));
  strcpy(cTemp,RecordBuffer);

  cToken = strtok(cTemp,cElementSeparatorString);
  for (i=1; i < 4; i++)
  {
    cToken = strtok(NULL,cElementSeparatorString);
    if ( (cToken == NULL) && (i < 3) )
    {
      cerr << INSUFFICIENT_ELEMENTS << endl
           << "Expected:  2  Found: " << i << " in ST"<< endl
           << RecordBuffer << endl;
      return FAILURE;
    }
    if (cToken != NULL)
    {
      strcpy(ST[i],cToken);
    }
  }
  return SUCCESS;
}

int X12RecordReader::parseSE()
{
  int i;
  char cTemp[32];
  char * cToken;
  memset(cTemp,0,sizeof(cTemp));
  strcpy(cTemp,RecordBuffer);

  cToken = strtok(cTemp,cElementSeparatorString);
  for (i=1; i < 3; i++)
  {
    cToken = strtok(NULL,cElementSeparatorString);
    if (cToken == NULL)
    {
      cerr << INSUFFICIENT_ELEMENTS << endl
           << "Expected:  2  Found: " << i << " in SE"<< endl
           << RecordBuffer << endl;
      return FAILURE;
    }
    strcpy(SE[i],cToken);
  }
  return SUCCESS;
}

int X12RecordReader::logST()
{
  cout << "    Processing Transaction Set\n"
    << "        Transaction Set ID:     " <<  ST[1] << endl
    << "        Control Number:         " <<  ST[2] << endl
    << endl;
  return SUCCESS;
}

int X12RecordReader::logSE()
{
  cout << "    Processed Transaction Set\n"
    << "        Included Segments:      " <<  SE[1] << endl
    << "        Control Number:         " <<  SE[2] << endl
    << endl;
  return SUCCESS;
}
