/* X12TargetConverter.cpp

Drives conversion from XML to an X12 EDI interchange

From Chapter 9 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include <strstrea.h>
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"


//  Special headers for this module only
#include "EDIRecordWriter.h"
#include "X12RecordWriter.h"
#include "X12TargetConverter.h"


//  Default Constructor
X12TargetConverter::X12TargetConverter(
     const char * const cFileDescriptionDocumentName,
     ofstream * const OutputStream) : 
     TargetConverter(cFileDescriptionDocumentName)
{
  //  call base class constructor, passing
  //    File Description Document Name

  //  Create X12RecordWriter object, passing:
  //    File Description Document DOM and output stream
  X12Writer = new X12RecordWriter(spFileDescription,
      OutputStream);
  Writer = X12Writer;
}


//  Destructor
X12TargetConverter::~X12TargetConverter()
{
  delete X12Writer;
}

//  Process a Document
int X12TargetConverter::processDocument(
    const IXMLDOMDocument2Ptr spDocInput)
{
  //  Locals
  IXMLDOMElementPtr spRootElement;
  IXMLDOMElementPtr spSegmentGrammar;
  IXMLDOMElementPtr spSegment;
  IXMLDOMNodePtr spGrammarChild;
  IXMLDOMNodePtr spRootChild;

  //  Basic types
  char cInputRootElementName[NMTOKEN_SIZE];
  char cGrammarElementName[32];
  char cGrammarName[NMTOKEN_SIZE];
  char cSegmentName[NMTOKEN_SIZE];

  int iReturn;

  memset(cInputRootElementName, 0,
      sizeof(cInputRootElementName));

  //  Root Element <- Get Document's documentElement attribute
  spRootElement = spDocInput->documentElement;

  //  Root Element Name <- Get Root Element's tagName attribute
  _bstr_t bstrInputRootElementName(spRootElement->tagName);

  //  IF Root Element Name != Grammar Root Element Name
  if ( !(strcmp(cRootElementName,
         (char *) bstrInputRootElementName) == 0) )
  {
    //  Return error
    cerr << bstrInputRootElementName << ROOT_MISMATCH 
         << cRootElementName << endl;
    return FAILURE;
  } //  ENDIF

  //  Call X12RecordWriter's writeST method to write
  //    Transaction Set Header segment
  X12Writer->writeST();

  //  ChildNode <- Get Grammar Element's firstChild
  spGrammarChild = spGrammar->firstChild;

  //  DO while ChildNode is not an Element Node
  while (spGrammarChild->nodeType != NODE_ELEMENT)
  {
    //  ChildNode <- nextSibling
    spGrammarChild = spGrammarChild->nextSibling;
  } //  ENDDO

  //  Segment Grammar Element <- ChildNode
  spSegmentGrammar = spGrammarChild;

  //  DO until all child Elements of the Root have been
  //    processed
  spRootChild = spRootElement->firstChild;
  while (spRootChild != NULL)
  {
    if (spRootChild->nodeType != NODE_ELEMENT)
    {
      spRootChild= spRootChild->nextSibling;
      continue;
    }
    spSegment = spRootChild;

    //  Segment Name <- call Segment Element's getNodeName
    memset(cSegmentName,0,sizeof(cSegmentName));
    strcpy(cSegmentName, _bstr_t(spSegment->nodeName));

    //  Grammar Name <- Call Segment Grammer getAttribute for
    //    "ElementName"
    memset(cGrammarName,0,sizeof(cGrammarName));
    strcpy(cGrammarName,
      _bstr_t(spSegmentGrammar->getAttribute("ElementName")));

    //  DO until Grammar Name = Segment Name
    while (strcmp(cGrammarName,cSegmentName))
    {
      //  Segment Grammar Element <- Get next Element
      //    from Grammar
      spGrammarChild= spGrammarChild->nextSibling;
      //  IF Record Grammar Element is NULL
      if (spGrammarChild == NULL)
      {
        //  Return error  - This record is not part of the group
        cerr << cSegmentName << RECORD_NOT_IN_GROUP_GRAMMAR
        << _bstr_t(spSegmentGrammar->getAttribute("ElementName"))
        << endl;
        return FAILURE;
      } //  ENDIF

      if (spGrammarChild ->nodeType != NODE_ELEMENT)
      {
        continue;
      }
      spSegmentGrammar = spGrammarChild;

      //  Grammar Tag = call Record Grammar Element's
      //    getAttribute for "ElementName"
      memset(cGrammarName,0,sizeof(cGrammarName));
      strcpy(cGrammarName,
          _bstr_t(
            spSegmentGrammar ->getAttribute("ElementName")));
    } //  ENDDO until Grammar Name = Name

    //  Grammar Element Name <- Call Record Grammar Element's
    //    getNodeName
    memset(cGrammarElementName,0,sizeof(cGrammarElementName));
    strcpy(cGrammarElementName,
        _bstr_t(spSegmentGrammar->nodeName));

    //  IF Grammar Element Name = "GroupDescription"
    if (!strcmp(cGrammarElementName,"GroupDescription"))
    {
      //  Call processGroup, passing Record Element and Record
      //    Grammar Element
      iReturn = processGroup(spSegment, spSegmentGrammar);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }
    }
    else  //  ELSE
    {
      //  Call RecordWriter's parseRecord, passing Record
      //    Element and Record Grammar
      iReturn = X12Writer->parseRecord(spSegment,
          spSegmentGrammar);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }
      //  Call writeRecord, passing Segment Grammar
      //    Element
      iReturn = X12Writer->writeRecord(spSegmentGrammar);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }
    } //  ENDIF

    //  Segment Elememt <- Get Segment Element's nextSibling,
    //    skipping over non-Element Nodes
    spRootChild = spRootChild->nextSibling;
  } //  END DO for all child Elements of Root

  // Call X12RecordWriter's writeSE method to write Transaction
  //    Set Trailer Segment
  X12Writer->writeSE();
  //  Return success
  return SUCCESS;
}

  //  Write interchange and group header segments
int X12TargetConverter::writeHeaders()
{
  int iReturn;
  iReturn = X12Writer->writeISA();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }
  return X12Writer->writeGS();
}

//  Write interchange and group trailer segments
int X12TargetConverter::writeTrailers()
{
  int iReturn;
  iReturn = X12Writer->writeGE();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }
  return X12Writer->writeIEA();
}
