/* DataCell.cpp

Base class for handling storage and conversion of field data

From Chapter 6 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Constructor with field number and grammar Element
DataCell::DataCell(int iFieldNumberIn,
  IXMLDOMElementPtr const spGrammarIn)
{
  iFieldNumber = iFieldNumberIn;
  iSubFieldNumber = 0;
  iBufferLength = CELL_BUFFER_SIZE;

  //  Initialize the array
  clearField();

  //  Save the grammar node
  spGrammar = spGrammarIn;
}


//  Destructor
DataCell::~DataCell() {};


//  Clear the field
void DataCell::clearField()
{
  memset(cCellBuffer,0,sizeof(cCellBuffer));
  iBufferLength = 0;
}

//  Get DelimitText Attribute of grammar node
bool DataCell::delimitText()
{
  char cDelimitText[6];

  //  Delimit Text <- call Grammar Element's getAttribute
  //    on "DelimitText"
  memset(cDelimitText,0,sizeof(cDelimitText));
  strcpy(cDelimitText, 
    _bstr_t(spGrammar->getAttribute("DelimitText")));

  //  IF Delimit Text String is false
  if ( !strcmp(cDelimitText,"false") )
  {
    //  return false
    return false;
  }  // ENDIF
  //  Return true
  return true;
}

//  Fill a field for output to legacy formats
void DataCell::fillField(const char cJustify,
                         const int iFixedLength)
{
  //  Locals
  char cMinimumLength[5];
  char cFillCharacter[2];
  int iLength;

  memset(cMinimumLength,0,sizeof(cMinimumLength));
  memset(cFillCharacter,0,sizeof(cFillCharacter));

  //  Length <- Fixed Length
  iLength = iFixedLength;

  //  IF ( Length = 0 )
  if (iLength == 0)
  {
    //  Minimum Length <- Call Grammar Element's getAttribute
    //    for "MinLength"
    strcpy(cMinimumLength,
      _bstr_t(spGrammar->getAttribute("MinLength")));
    //  Length <- Minimum Length, converted from string
    iLength = atoi(cMinimumLength);
  } //  ENDIF

  //  IF ( Cell Buffer Length >= Length )
  if (iBufferLength >= iLength)
  {
    return;
  } //  ENDIF

  //  Fill Character <- Call Grammar Element's getAttribute for
  //    "FillCharacter"
  _variant_t vFillCharacter(
      spGrammar->getAttribute("FillCharacter"));
  if (vFillCharacter.vt != VT_NULL)
  {
    strcpy(cFillCharacter, _bstr_t(vFillCharacter));
  }
  else
  {
    strcpy(cFillCharacter," ");
  }

  //  IF ( Justify = R )
  if (cJustify == 'R')
  {
    //  Call justifyFieldRight, passing Length and Fill
    //    Character
    justifyFieldRight(iLength,cFillCharacter[0]);
  }
  else  //  ELSE
  {
    //  Call justifyFieldLeft, passing Length and Fill
    //    Character
    justifyFieldLeft(iLength,cFillCharacter[0]);
  } //  ENDIF

  //  Return success
}

//  Return the Buffer Contents
int DataCell::getField(char * const cFieldContents)
{
  memcpy(cFieldContents,cCellBuffer,iBufferLength);
  return iBufferLength;
}

//  Return the field number
int DataCell::getFieldNumber()
{
  return iFieldNumber;
}

//  Return the subfield number
int DataCell::getSubFieldNumber()
{
  return iSubFieldNumber;
}


//  Return the length of the buffer
int DataCell::getLength()
{
  return iBufferLength;
}

//  Return the Element Name from the Grammar
void DataCell::getName(char * const cElementName)
{
  strcpy(cElementName,
    _bstr_t(spGrammar->getAttribute("ElementName")));
}

//  Left justify the Cell Buffer contents
//  Algorithm is dependent on language features
//  This method developed in Chapter 8
void DataCell::justifyFieldLeft(const int iLength,
                           const char cFill)
{
  //  Fill to right
  memset(&(cCellBuffer[iBufferLength]),cFill,
      (iLength - iBufferLength));
  iBufferLength = iLength;
}

//  Right justify the Cell Buffer contents
//  Algorithm is dependent on language features
//  This method developed in Chapter 8
void DataCell::justifyFieldRight(const int iLength,
                            const char cFill)
{
  int iFillLength;
  iFillLength = iLength - iBufferLength;
  //  Shift current contents to right
  memcpy(&(cCellBuffer[iFillLength]),cCellBuffer, iBufferLength);
  //  Fill left part
  memset(cCellBuffer,cFill,iFillLength);
  iBufferLength = iLength;
}

//  Append a byte to the field
int DataCell::putByte(char cFieldByte)
{
  //  Test for overflow
  if (iBufferLength == CELL_BUFFER_SIZE)
  {
    cerr << BUFFER_OVERFLOW_IN << "DataCell::putByte" << endl;
    return FAILURE;
  }
  cCellBuffer[iBufferLength] = cFieldByte;
  iBufferLength++;
  return SUCCESS;
}

//  Load the field
int DataCell::putField(const char *  const cFieldContents,
     int iFieldLength)
{
  //  Test for overflow
  if (iFieldLength > CELL_BUFFER_SIZE)
  {
    cerr <<  BUFFER_OVERFLOW_IN << "DataCell::putField" << endl;
    return FAILURE;
  }
  
  //  Clear it
  clearField();

  //  Load it
  memcpy(cCellBuffer, cFieldContents, iFieldLength);
  iBufferLength = iFieldLength;

  return SUCCESS;
}

//  Set the subfield number
void DataCell::setSubFieldNumber(int iSubFieldNumberIn)
{
  iSubFieldNumber = iSubFieldNumberIn;
}


//  Write an Element from the cell
void DataCell::toElement(const IXMLDOMDocument2Ptr spDocOutput,
    const IXMLDOMElementPtr spParent)
{
  //  Locals
  char cElementName[NMTOKEN_SIZE];

  //  IF BufferLength = 0
  if (iBufferLength == 0)
  {
    //  Return
    return;
  } //  ENDIF
  
  //  Get the Element name from the grammar node
  memset(cElementName,0,sizeof(cElementName));
  getName(cElementName);

  //  Create the Element in the output document
  IXMLDOMElementPtr spNewElement = 
        spDocOutput->createElement(cElementName);

  //  Append it to the parent
  spParent->appendChild(spNewElement);

  //  Create the text node
  IXMLDOMTextPtr spTexText = spDocOutput->createTextNode(
        cCellBuffer);

  //  Append it to the new element
  spNewElement->appendChild(spTexText);

}

//  Remove leading and trailing white space from Cell Buffer
//  contents - All characters with an int value less than
//  or equal to a space are trimmed
void DataCell::trim()
{
  //  Locals
  int iFirst, iLen;

  //  Trim trailing white space
  while ( (iBufferLength > 0) &&
    ( cCellBuffer[iBufferLength-1] <= ' ') )
  {
    cCellBuffer[--iBufferLength] = 0;
  }

  //  Exit if we have an empty string
  if (iBufferLength == 0)
  {
    return;
  }

  //  Find first non whitespace character
  iFirst = 0;
  while ( (cCellBuffer[iFirst] <= ' ') &&
    (iFirst < iBufferLength) )
  {
    iFirst++;
  }

  //  Trim by shifting if required
  //  We can use strcpy since the data is string by now
  if (iFirst > 0)
  {
    iLen = iBufferLength - iFirst;
    strcpy(cCellBuffer, &(cCellBuffer[iFirst]));
  }

  iBufferLength = strlen(cCellBuffer);
}


//  Trim leading zeroes and white space from numeric fields
//  We support the following numeric formats
//    Sign character in first position followed by leading
//      spaces or zeroes
//    Leading spaces followed by leading sign character
//    Leading spaces or zeroes, with no sign character

void DataCell::trimLeadingZeroes()
{

  //  Locals
  char cTarget[CELL_BUFFER_SIZE];
  int iPos, iLen;
  int iStart = 0;

  //  Initialize Sign Character
  char cSignChar = 0;

  //  IF first character is + or - sign character
  if ( (cCellBuffer[0] == '-') ||
       (cCellBuffer[0] == '+') )
  {
    //  Sign Character <- first Character
    cSignChar = cCellBuffer[0];
    iStart = 1;
  } //  ENDIF

  //  Call trim (C++ base class method or native Java) to
  //  remove leading and trailing white space from CellBuffer
  //  For C++ we first skip over the sign character
  if ( cSignChar != 0 )
  {
    strcpy(cCellBuffer,&(cCellBuffer[1]));
  }

  trim();

  //  Clear buffer and exit if empty after trimming
  if (iBufferLength == 0)
  {
    return;
  }

  //  Position <- 0
  iPos = 0;

  //  IF first character is + or - sign character
  if ( (cCellBuffer[0] == '-') ||
       (cCellBuffer[0] == '+') )
  {
    //  Sign Character <- first Character
    cSignChar = cCellBuffer[0];
    iPos = 1;
  } //  ENDIF

  iLen = strlen(cCellBuffer);

  //  DO while Position < CellBuffer length and
  //    CellBuffer[Position] = "0"
  while ( (iPos < iLen ) &&
          (cCellBuffer[iPos] == '0') )
  {
    //  Position++
    iPos++;
  } //  ENDDO

  //  IF Position = CellBuffer length because it contains only
  //    zeroes
  if (iPos == iLen)
  {
    //  Decrement Position so that we have at least one zero
    iPos--;
  } //  ENDIF

  memset(cTarget,0,sizeof(cTarget));
  //  IF Sign Character is present
  if (cSignChar != 0)
  {
    //  CellBuffer <- Sign Character + CellBuffer substring
    //    starting at Position
    cTarget[0] = cSignChar;
    strcat(cTarget,&(cCellBuffer[iPos]));
  }
  else  //  ELSE
  {
    //  CellBuffer <- CellBuffer substring starting at Position
    strcpy(cTarget,&(cCellBuffer[iPos]));
  } //  ENDIF

  //  Save it
  putField(cTarget,strlen(cTarget));
  //  Return success
}


//  Stub functions for derived classes
int DataCell::fromXML() { return SUCCESS; }
int DataCell::prepareOutput() { return SUCCESS; }
int DataCell::toXML() { return SUCCESS; }



