/* DataCellReal.cpp

For handling of Real (decimal) fields in non-XML files

From Chapter 7 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Constructor with field number and grammar Element
DataCellReal::DataCellReal(int iFieldNumberIn,
    IXMLDOMElementPtr spGrammarIn) :
  DataCell(iFieldNumberIn, spGrammarIn) { }

//  Destructor
DataCellReal::~DataCellReal() {};


//  Convert from XML Schema decimal to real
//  Trim spaces - If schema validation is done we won't have
//  spaces in a decimal field, but we're not going to
//  depend on them doing validation
int DataCellReal::fromXML()
{
  //  Trim whitespace
  trim();

  //  Remove leading plus sign.  XML types allow it, but some
  //  legacy formats such as X12 don't; it is always implied.
  if ( cCellBuffer[0] == '+' )
  {
    strcpy(cCellBuffer,&(cCellBuffer[1]));
    cCellBuffer[iBufferLength] = 0;
    iBufferLength--;
  }

  
  return SUCCESS;
}

//  Prepare for output to legacy formats
//  Added in Chapter 8
int DataCellReal::prepareOutput()
{
  char cLength[5];
  char cTruncate[8];
  char cSignCharacter;
  bool boDecimalTrucated = false;
  int i, iLength = 0;
  int iMinLength = 0;
  _variant_t vLength;

  memset(cLength,0,sizeof(cLength));
  memset(cTruncate,0,sizeof(cTruncate));

  memset(cLength,0,sizeof(cLength));
  memset(cTruncate,0,sizeof(cTruncate));

  //  Length <- Call Grammar Element's getAttribute for
  //    "Length"
  //  We may have Length or MinLength - With MSXML we get
  //  a variant and have to check if it is null
  vLength = spGrammar->getAttribute("Length");
  if (vLength.vt != VT_NULL)
  {
    strcpy(cLength,_bstr_t(vLength));
    if (strlen(cLength) > 0)
    {
      iLength = atoi(cLength);
      iMinLength = iLength;
    }
  }
  else
  {
    strcpy(cLength,
      _bstr_t(spGrammar->getAttribute("MinLength")));
    iMinLength = atoi(cLength);
  }

  //  Sign Character <- First character in Cell Buffer
  cSignCharacter = cCellBuffer[0];

  //  Call fillField, passing R for right justification, and
  //    Length
  fillField('R', iMinLength);

  //  IF (Sign Character = "+" or "-") and (First character in
  //    Cell Buffer = "0")
  if ( ((cSignCharacter == '+') || (cSignCharacter == '-')) &&
       (cCellBuffer[0] == '0') )
  {
    //  Dependent on language features, find Sign Character in
    //  Cell Buffer and swap with first character in Cell
    //  Buffer
    i = 1;
    while (cCellBuffer[i] != cSignCharacter)
    {
      i++;
    }
    cCellBuffer[0] = cSignCharacter;
    cCellBuffer[i] = '0';
  } //  ENDIF

  //  IF ( Length = 0 )
  if (iLength == 0)
  {
    //  Maximum Length <- Call Grammar Element's getAttribute
    //    for "MaxLength"
    memset(cLength,0,sizeof(cLength));
    strcpy(cLength,
        _bstr_t(spGrammar->getAttribute("MaxLength")));
    //  Length <- Maximum Length, converted from string
    iLength = atoi(cLength);
  } //  ENDIF

  //  IF (Cell Buffer Length <= Length)
  if (iBufferLength <= iLength)
  {
    //  Return success
    return SUCCESS;
  } //  ENDIF

  //  Truncate <- Call Grammar Element's getAttribute for
  //    "Truncatable"
  strcpy(cTruncate,
      _bstr_t(spGrammar->getAttribute("Truncatable")));

  //  IF ( Truncate == false )
  if (!strcmp(cTruncate,"false"))
  {
    //  Return error
    cerr << NO_TRUNCATION << endl
         << _bstr_t(spGrammar->getAttribute("ElementName"))
         <<  " " << iLength << endl
         << cCellBuffer << endl;
    return FAILURE;
  } //  ENDIF

  //  Depending on laguage features, truncate Cell Buffer
  //  contents to specified Length stopping when Length is
  //  achieved or decimal point is truncated
  //  First we assume that there is not decimal point in the
  //    number, then we look for it.
  boDecimalTrucated = true;
  for (i=0; i < iBufferLength; i++)
  {
    if (cCellBuffer[i] == '.')
    {
      boDecimalTrucated = false;
    }
  }

  while ( (iBufferLength > iLength) && !boDecimalTrucated)
  {
    iBufferLength--;
    if (cCellBuffer[iBufferLength] == '.')
    {
      boDecimalTrucated = true;
    }
    cCellBuffer[iBufferLength] = 0;
  }

  //  IF (Cell Buffer Length > Length)
  if (iBufferLength > iLength)
  {
    //  Return failure
    cerr << NO_TRUNCATION << endl
         << _bstr_t(spGrammar->getAttribute("ElementName"))
         <<  " " << iLength << endl
         << cCellBuffer << endl;
    return FAILURE;
  } //  ENDIF

  //  Cell Buffer Length <- Length
  //  Return success

  return SUCCESS;
}



//  Convert Real to XML decimal type
int DataCellReal::toXML()
{
  //  Trim whitespace and leading zeroes
  trimLeadingZeroes();
  return SUCCESS;
}
