/* DataCellX12R.cpp

DataCell derived class for X12 Decimal data type

From Chapter 9 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Constructor with field number and grammar Element
DataCellX12R::DataCellX12R(int iFieldNumberIn,
    IXMLDOMElementPtr spGrammarIn) :
  DataCellReal(iFieldNumberIn, spGrammarIn) { }

//  Destructor
DataCellX12R::~DataCellX12R() {};


//  Convert from XML Schema decimal to real
//  Trim spaces - If schema validation is done we won't have
//  spaces in a decimal field, but we're not going to
//  depend on them doing validation
int DataCellX12R::fromXML()
{
  char * ptr;
  //  Trim whitespace
  trim();


  //  IF CellBuffer = -0, INF, -INF, or NaN
  if ( !strcmp(cCellBuffer,"-0") || !strcmp(cCellBuffer,"INF") ||
       !strcmp(cCellBuffer, "-INF") || !strcmp(cCellBuffer,"NaN") )
  {
    //  Return error for unsupported constant in X12 Decimal
    //    data type
    cerr << UNSUPPORTED_SPECIAL_VALUE << " " 
         << _bstr_t(spGrammar->getAttribute("ElementName"))
         << " X12-R " << cCellBuffer << endl;
  } //  ENDIF

  //  Remove leading plus sign since X12 doesn't allow it.
  if (cCellBuffer[0] == '+')
  {
    strcpy(cCellBuffer,&(cCellBuffer[1]));
    iBufferLength--;
    cCellBuffer[iBufferLength] = 0;
  }

  //  Convert to uppercase since XML schema datatypes allow a
  //    lower case 'e' to indicate the exponent
  ptr = strchr(cCellBuffer,'e');
  if (ptr != NULL)
  {
    *ptr = 'E';
  }

  return SUCCESS;
}

int DataCellX12R::prepareOutput()
{
  char cMinLength[5], cMaxLength[5];
  char cTruncate[8];
  int iMinLength, iMaxLength;
  //  NegativeNumber <- false
  bool boNegativeNumber = false;
  bool boExponentialNotation = false;
  bool boDecimalTrucated;

  memset(cMinLength,0,sizeof(cMinLength));
  memset(cMaxLength,0,sizeof(cMaxLength));
  memset(cTruncate,0,sizeof(cTruncate));

  //  IF first character in CellBuffer is "-"
  if (cCellBuffer[0] == '-')
  {
    //  CellBuffer <- Shift CellBuffer left one character to
    //  remove Minus sign
    strcpy(cCellBuffer,&(cCellBuffer[1]));
    iBufferLength--;
    cCellBuffer[iBufferLength] = 0;
    //  NegativeNumber <- true
    boNegativeNumber = true;
  } //  ENDIF

  //  Minimum Length <- Call Grammar Element's getAttribute for
  //    "MinLength"
  strcpy(cMinLength, _bstr_t(spGrammar->getAttribute("MinLength")));
  iMinLength = atoi(cMinLength);

  //  Maximum Length <- Call Grammar Element's getAttribute
  //    for "MaxLength"
  strcpy(cMaxLength,
      _bstr_t(spGrammar->getAttribute("MaxLength")));
  //  Length <- Maximum Length, converted from string
  iMaxLength = atoi(cMaxLength);

  //  IF Cell Buffer contains "E"
  if (strchr(cCellBuffer,'E') != NULL)
  {
    //  Increment Minimum and Maximum Length
    iMinLength++;
    iMaxLength++;
    boExponentialNotation = true;
  } //  ENDIF

    //  IF Cell Buffer contains a decimal point
  boDecimalTrucated = true;
  if (strchr(cCellBuffer,'.') != NULL)
  {
    //  Increment Minimum and Maximum Length
    iMinLength++;
    iMaxLength++;
    boDecimalTrucated = false;
  } //  ENDIF

  //  IF ( Cell Buffer Length < Minimum Length )
  //  Call justifyFieldRight, passing Minimum Length and zero
  //    as Fill Character
  //  ENDIF
  if (iBufferLength < iMinLength)
  {
    justifyFieldRight(iMinLength,'0');
  }

  //  IF (Cell Buffer Length > Maximum Length)
  if (iBufferLength > iMaxLength)
  {
    //  Truncate <- Call Grammar Element's getAttribute for
    //    "Truncatable"
    strcpy(cTruncate,
        _bstr_t(spGrammar->getAttribute("Truncatable")));

    //  IF ( Truncate == false ) OR ExponentialNotation = true
    if (!strcmp(cTruncate,"false") || boExponentialNotation)
    {
      //  Return error
      cerr << NO_TRUNCATION << endl
           << _bstr_t(spGrammar->getAttribute("ElementName"))
           <<  " " << iMaxLength << endl
           << cCellBuffer << endl;
      return FAILURE;
    } //  ENDIF

    //  Depending on laguage features, truncate Cell Buffer
    //  contents to specified Length stopping when Length is
    //  achieved or decimal point is truncated
    while ( (iBufferLength > iMaxLength) && !boDecimalTrucated)
    {
      iBufferLength--;
      if (cCellBuffer[iBufferLength] == '.')
      {
        boDecimalTrucated = true;
      }
      cCellBuffer[iBufferLength] = 0;
    }

    //  IF (Cell Buffer Length > Length)
    if (iBufferLength > iMaxLength)
    {
      //  Return failure
      cerr << NO_TRUNCATION << endl
           << _bstr_t(spGrammar->getAttribute("ElementName"))
           <<  " " << iMaxLength << endl
           << cCellBuffer << endl;
      return FAILURE;
    } //  ENDIF
  }
  //  IF NegativeNumber = true
  if (boNegativeNumber)
  {
    //  CellBuffer <- Shift CellBuffer right one character to
    //    insert Minus sign
    //  Increment Cell Buffer Length
    strcpy(&(cCellBuffer[1]),cCellBuffer);
    cCellBuffer[0] = '-';
    iBufferLength++;
  } //  ENDIF
  //  Cell Buffer Length <- Length
  //  Return success

  return SUCCESS;
}

//  Convert toXML uses base class method
