/* RecordHandler.cpp

Base class for handling records in non-XML formats

From Chapter 6 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"


//  Special headers for this module only
#include <strstrea.h>


RecordHandler::RecordHandler(const IXMLDOMDocument2Ptr 
                             spFileDescriptionIn)
{
  //  Locals
  int i;
  
  //  Initialize the array
  for (i = 0; i < CELL_ARRAY_SIZE; i++ )
  {
    CellArray[i] = NULL;
  }

  spFileDescription = spFileDescriptionIn;
  iRecordBufferLength = 0;
  iHighestCell = -1;
}


//  Destructor
RecordHandler::~RecordHandler() {};


int RecordHandler::getElementText(
    const IXMLDOMElementPtr spElementIn,
    char * const cElementText,
    int iBufferLength)
{ 
  //  Assume the text node is the first child
  IXMLDOMNodePtr spChildNode = spElementIn->firstChild;
  memset(cElementText,0,iBufferLength);

  //  Check, and keep looking until we find it
  while ( (spChildNode != NULL) &&
    (spChildNode->nodeType != NODE_TEXT) )
  {
    spChildNode = spChildNode->nextSibling;
  }

  //  Return success with no action if no text child found
  if (spChildNode == NULL)
  {
    return SUCCESS;
  }

  _bstr_t bstrNodeValue(spChildNode->nodeValue);
  //  Test for overflow before copying
  if (bstrNodeValue.length() > (unsigned int) iBufferLength)
  {
    cerr << BUFFER_OVERFLOW_IN << "RecordHandler::getElementText"
      << endl;
    return FAILURE;
  }

  strcpy(cElementText, bstrNodeValue);
  return SUCCESS;
}


//  Utility method to get the contents of an entry in the
//  DataCell Arrray
//  Returns 0 on success, 1 on failure
int RecordHandler::getFieldValue(int iFieldNumber,
                                 char * const cFieldContents)
{ 
  int i;

  //  Run through DataCell Array to get requested value
  for (i = 0; i <= iHighestCell; i++)
  {
    if (iFieldNumber == CellArray[i]->getFieldNumber())
    {
      CellArray[i]->getField(cFieldContents);
      return SUCCESS;
    }
  }

  //  Return error if we can't find it
  cerr << NO_FIELD << iFieldNumber << endl;
  return FAILURE;
}


char RecordHandler::setDelimiter(const char * const cDelimiter)
{

  //  IF length of Delimiter = 1
  if (strlen(cDelimiter) == 1)
  {
    //  return Delimiter as byte
    return cDelimiter[0];
  } //  ENDIF

  //  return Delimiter converted to byte from hex string
  istrstream HexIn((char *) cDelimiter, strlen(cDelimiter));
  HexIn.setf(ios::basefield,ios::hex);

  int iTemp;

  //  Read hex value from string stream
  HexIn >> iTemp;
  return iTemp;
}

void RecordHandler::setTerminator(const char * const cTerminator)
{
  //  IF length of Terminator = 1
  if (strlen(cTerminator) == 1)
  {
    //  DO CASE of Terminator
    switch(cTerminator[0])
    {
      //  'U':
      case 'U':
        //  Record Terminator1 = line feed
        //  BREAK:
        cRecordTerminator1 = 10;
        cRecordTerminator2 = 0;
        break;
      //  'W':
      case 'W':
        //  Record Terminator1 = carriage return
        //  Record Terminator2 = line feed
        //  BREAK:
        cRecordTerminator1 = 13;
        cRecordTerminator2 = 10;
        break;
      //    other:
      default:
        //  Record Terminator1 = passed Terminator
        cRecordTerminator1 = cTerminator[0];
        cRecordTerminator2 = 0;
    } //  ENDDO
  }
  else  //ELSE
  {
    //  Set Terminator1 from hexadecimal value
    istrstream HexIn((char *) cTerminator, strlen(cTerminator));
    HexIn.setf(ios::basefield,ios::hex);

    int iTemp;

    //  Read hex value from string stream
    HexIn >> iTemp;
      
    cRecordTerminator1 = iTemp;
    cRecordTerminator2 = 0;
  } //ENDIF
}


//  Create DataCell of requested type
//  This is where the list of new data types and codes
//  is maintained
//  EDIT THIS WHEN ADDING NEW DATACELL DERIVED CLASSES
int RecordHandler::createDataCell(int iFieldNumber,
    const IXMLDOMElementPtr spGrammar, 
    DataCell * &NewCell)
{
  //  Locals
  //  XML

  //  COM

  //  Project

  
  //  Basic
  char cCellDataType[32];

  //  Test for Array overflow
  if (iHighestCell == CELL_ARRAY_SIZE)
  {
    cerr << ARRAY_OVERFLOW_IN << "RecordHandler::createDataCell"
      << endl;
    return FAILURE;
  }
  
  //  Initializations
  NewCell = NULL;
  
  //  Get the data type from the grammar Element
  memset(cCellDataType,'\0',sizeof(cCellDataType));
  strcpy(cCellDataType, 
    _bstr_t(spGrammar->getAttribute("DataType")) );

 
  //  Create cell based on passed type in Grammar Element

  //  Alphanumeric string, X12 AN, or X12 ID
  if ( (strcmp(cCellDataType,"AN") == 0) ||
       (strcmp(cCellDataType,"X12-AN") == 0) ||
       (strcmp(cCellDataType,"X12-ID") == 0) )
  {
    NewCell = new DataCellAN(iFieldNumber, spGrammar);
  }

  //  Real
  if (strcmp(cCellDataType,"R") == 0)
  {
    NewCell = new DataCellReal(iFieldNumber, spGrammar);
  }

  //  Numeric with implied decimal
  if (cCellDataType[0] == 'N')
  {
    NewCell = new DataCellN(iFieldNumber, spGrammar);
  }

  //  Date in MM/DD/YYYY
  if (strcmp(cCellDataType,"DMMsDDsYYYY") == 0)
  {
    NewCell =
      new DataCellDateMMsDDsYYYY(iFieldNumber, spGrammar);
  }

  //  Date in YYYYMMDD
  if (strcmp(cCellDataType,"DYYYYMMDD") == 0)
  {
    NewCell =
      new DataCellDateYYYYMMDD(iFieldNumber, spGrammar);
  }

  //  X12 EDI Date Type
  if (strcmp(cCellDataType,"X12-DT") == 0)
  {
    NewCell =
      new DataCellX12DT(iFieldNumber, spGrammar);
  }

  //  X12 EDI Time Type
  if (strcmp(cCellDataType,"X12-TM") == 0)
  {
    NewCell =
      new DataCellX12TM(iFieldNumber, spGrammar);
  }
  
  //  X12 EDI Numeric Type
  if (strncmp(cCellDataType,"X12-N", 5) == 0)
  {
    NewCell =
      new DataCellX12N(iFieldNumber, spGrammar);
  }
  
  //  X12 EDI Real Type
  if (strcmp(cCellDataType,"X12-R") == 0)
  {
    NewCell =
      new DataCellX12R(iFieldNumber, spGrammar);
  }
  
  //  Insert new if blocks for new cell type above this line
  if (NewCell != NULL)
  {
    iHighestCell++;
    CellArray[iHighestCell] = NewCell;
    return SUCCESS;
  }

  //  Return error for unsupported type
  cerr << UNSUPPORTED_DATA_TYPE << cCellDataType << endl;
  return FAILURE;

}  

