/* RecordReader.cpp

Base class for reading records in non-XML formats

From Chapter 6 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"


//  Constructor with terminator only
RecordReader::RecordReader(
    const IXMLDOMDocument2Ptr spFileDescription) :
    RecordHandler(spFileDescription) { }

//  Constructor with output stream and terminator
RecordReader::RecordReader(
    const IXMLDOMDocument2Ptr spFileDescription,
    ifstream * const InputFileIn) :
  RecordHandler(spFileDescription)
{
  //  Save output stream
  InputFile = InputFileIn;
}

//  Destructor
RecordReader::~RecordReader() {};


//  Read a variable length record from the input file
//  Return length, IO_READ_EOF if EOF, 
//  and IO_READ_ERROR if error.
//  Under normal circumstances the last record in the file will
//  end with the expected terminator(s).  However, if it doesn't
//  we'll return what we read, then return an EOF on the
//  next read.
int RecordReader::readRecordVariableLength()
{
  iRecordBufferLength = 0;
  memset(RecordBuffer,'\0',sizeof(RecordBuffer));

  //  Single terminator
  if (cRecordTerminator2 == 0)
  {
    //  Read until we hit the terminator or end of file
    //  At EOF we will get -1 as the byte read
    do
    {
      //  Test for buffer overflow before reading
      if (iRecordBufferLength == MAX_RECORD_SIZE)
      {
        cerr << BUFFER_OVERFLOW_IN << 
          "RecordReader::readRecordVariableLength" << endl;
        return IO_OTHER_ERROR;
      }
      InputFile->get(RecordBuffer[iRecordBufferLength]);
	    if (InputFile->bad())
      {
        return IO_READ_ERROR;
      }
      iRecordBufferLength++;

    } while ( (RecordBuffer[iRecordBufferLength - 1]
      != cRecordTerminator1) && !InputFile->eof());

    //  Back off the last character since we don't want to
    //  return either the terminator or the eof character
    RecordBuffer[iRecordBufferLength - 1] = 0;
    iRecordBufferLength--;

  }
  else
  //  Two terminators - probably CR and LF
  {
    //  Read until we hit the terminators or end of file
    bool boHaveRecord = false;
    do
    {
      //  Test for buffer overflow before reading
      if (iRecordBufferLength == MAX_RECORD_SIZE)
      {
        cerr << BUFFER_OVERFLOW_IN << 
          "RecordReader::readRecordVariableLength" << endl;
        return IO_OTHER_ERROR;
      }
      InputFile->get(RecordBuffer[iRecordBufferLength]);
	    if (InputFile->bad())
      {
        return IO_READ_ERROR;
      }

      iRecordBufferLength++;

      //  Make sure we have at least two bytes and the
      //  last two are the terminators,
      //  or that we are at EOF
      if ( ((iRecordBufferLength >= 2) &&
        (RecordBuffer[iRecordBufferLength-2] == cRecordTerminator1) &&
        (RecordBuffer[iRecordBufferLength-1] == cRecordTerminator2) )
          || InputFile->eof() )
      {
        boHaveRecord = true;
      }

    } while ( !boHaveRecord );

    //  Back off the terminators.
    //  We'll have at least one for the -1 for EOF
    RecordBuffer[iRecordBufferLength - 1] = 0;
    iRecordBufferLength--;

    //  Back off the first terminator if present
    if ( (iRecordBufferLength >= 1) && 
         (RecordBuffer[iRecordBufferLength - 1] == cRecordTerminator1) )
    {
      RecordBuffer[iRecordBufferLength - 1] = 0;
      iRecordBufferLength--;
    }
  } //  End else for two terminators

  //  Return EOF if we have it and nothing else
  if ( InputFile->eof() && (iRecordBufferLength == 0) )
  {
    return IO_READ_EOF;
  }
  //  Otherwise return the record length  
  else
  {
    return iRecordBufferLength;
  }
}

//  Set the current output document
void RecordReader::setOutputDocument(
     const IXMLDOMDocument2Ptr spDocOutputIn)
{
  spDocOutput = spDocOutputIn;
}


//  Set the input stream
void RecordReader::setInputStream(ifstream * InputStream)
{
  InputFile = InputStream;
}


//  Convert DataCell Array cells to XML format
int RecordReader::toXMLType()
{
  int i, iReturn;
  for (i = 0; i <= iHighestCell; i++)
  {
    iReturn = CellArray[i]->toXML();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
  }
  return SUCCESS;
}


//  Dummy base class method - Returns error
int RecordReader::getRecordType(char * const cRecordType)
{
  return FAILURE;
}

//  Dummy base class method - Returns error
int RecordReader::parseRecord(
    const IXMLDOMElementPtr spRecordGrammar)
{
  return FAILURE;
}


//  Dummy base class method - Returns error
int RecordReader::readRecord()
{
  return FAILURE;
}

//  Write a Record to the output XML Document from the
//  DataCell Array
void RecordReader::writeRecord(const IXMLDOMElementPtr spParent,
    const IXMLDOMElementPtr spRecordGrammar)
{
  //  Locals

  //  COM

  //  Basic
  char cElementName[NMTOKEN_SIZE];
  int i;

  //  Record Element <- call Output Document's createElement
  memset(cElementName,0,sizeof(cElementName));
  strcpy(cElementName,
      _bstr_t(spRecordGrammar->getAttribute("ElementName")));
  
  IXMLDOMElementPtr spRecordElement =
    spDocOutput->createElement(cElementName);

  //  Parent Element <- call Parent's appendChild to append
  //    Record Element
  spParent->appendChild(spRecordElement);

  //  DO for all DataCells in array up to Highest Cell
  for (i = 0; i <= iHighestCell; i++)
  {
    //  Call toElement on DataCell to create Element, load
    //  the text and attach it to the parent
    CellArray[i]->toElement(spDocOutput,spRecordElement);

    //  Clear Cell Array Entry
    delete CellArray[i];

  } //  ENDDO

  //  Clear highest cell back to -1
  iHighestCell = -1;
}



