/* RecordWriter.cpp

Base class for writing records to non-XML formats

From Chapter 6 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"


//  Basic Constructor 
RecordWriter::RecordWriter(
  const IXMLDOMDocument2Ptr spFileDescription) :
  RecordHandler(spFileDescription)
{
  //  Do nothing here
}

//  Constructor with output stream
RecordWriter::RecordWriter(
  const IXMLDOMDocument2Ptr spFileDescription,
  ofstream * const OutputFileIn) :
  RecordHandler(spFileDescription)
{
  //  Save output stream
  OutputFile = OutputFileIn;
}

//  Destructor
RecordWriter::~RecordWriter() {};


//  Process a record from the input XML document into the
//  DataCell Array
//  This is used for CSV and Flat File formats
int RecordWriter::parseRecord(
    const IXMLDOMElementPtr spRecordElement,
    const IXMLDOMElementPtr spRecordGrammar)
{
  //  XML
  IXMLDOMElementPtr spFieldElement;
  IXMLDOMNodePtr spChildNode;
  IXMLDOMElementPtr spFieldElementGrammar;
  IXMLDOMNodePtr spChildNodeGrammar;

  //  COM

  //  Project
  DataCell * NewCell = NULL;

  //  Basic
  char cRecordElementName[NMTOKEN_SIZE];
  char cGrammarFieldName[NMTOKEN_SIZE];

  char cFieldNumber[4];
  char cSubFieldNumber[4];
  char cElementText[CELL_BUFFER_SIZE];
  int iFieldNumber, iSubFieldNumber, iReturn;

  //  Field Grammar <- Get first Grammar child Element from
  //  Record Grammar Element, skipping over non-Element nodes
  spChildNodeGrammar = spRecordGrammar->firstChild;

  while (spChildNodeGrammar->nodeType != NODE_ELEMENT)
  {
    spChildNodeGrammar = spChildNodeGrammar->nextSibling;
  }
  spFieldElementGrammar = spChildNodeGrammar;

  //  Grammar Field Name <- call Field Grammar's getAttribute
  //  on "ElementName"
  memset(cGrammarFieldName,0,sizeof(cGrammarFieldName));
  strcpy(cGrammarFieldName,
    _bstr_t(spFieldElementGrammar->getAttribute("ElementName")));

  //  Get the first Field child Element of the data Record
  spChildNode = spRecordElement->firstChild;

  //  DO for all Field Elements in data Record's NodeList
  while (spChildNode != NULL)
  {
    //  Skip any non-Element Nodes
    if (spChildNode->nodeType != NODE_ELEMENT)
    {
      spChildNode = spChildNode->nextSibling;
      continue;
    }

    spFieldElement = spChildNode;

    //  ElementText <- call getElement Text on Field Element
    iReturn = getElementText(spFieldElement, cElementText,
      sizeof(cElementText));
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  IF Element Text is empty
    if ( strlen(cElementText) == 0)
    {
      //  Proceed to next field
      spChildNode = spChildNode->nextSibling;
      continue;
    } //  ENDIF
   
    //  Field Name <- Field Element's tagName attribute
    _bstr_t bstrTagName(spFieldElement->tagName);

    //  DO UNTIL Field Name == Grammar Field Name
    while (strcmp(cGrammarFieldName,bstrTagName))
    {
      //  Get next sibling Field Grammar Element
      spChildNodeGrammar = spChildNodeGrammar->nextSibling;

      //  Skip over non-Element Nodes
      while ( ( spChildNodeGrammar != NULL) &&
      (spChildNodeGrammar->nodeType != NODE_ELEMENT) )
      {
        spChildNodeGrammar = spChildNodeGrammar->nextSibling;
      }

      if (spChildNodeGrammar == NULL)
      {
        // Return error for Element not in Record Grammar
        memset(cRecordElementName,0,sizeof(cRecordElementName));
        strcpy(cRecordElementName,
            _bstr_t(spRecordGrammar->getAttribute("ElementName")));
        cerr << bstrTagName
             << FIELD_ELEMENT_NOT_IN_RECORD_GRAMMAR
             << cRecordElementName << endl;
        return FAILURE;
      }

      spFieldElementGrammar = spChildNodeGrammar;

      //  Grammar Field Name <- call Field Grammar's
      //    getAttribute on "ElementName"
      memset(cGrammarFieldName,0,sizeof(cGrammarFieldName));
      strcpy(cGrammarFieldName,
        _bstr_t(spFieldElementGrammar->getAttribute("ElementName")));
    }

    //  Field Number <- Get call Field Grammar's getAttribute
    //    on "Number"
    memset(cFieldNumber,0,sizeof(cFieldNumber));
    strcpy(cFieldNumber,
    _bstr_t(spFieldElementGrammar->getAttribute("FieldNumber")));
    iFieldNumber = atoi(cFieldNumber);

    //  NewCell <- call createDataCell, passing Field Number
    //    and DataType
    iReturn = createDataCell(iFieldNumber, 
      spFieldElementGrammar, NewCell);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  SubField Number <- Get call Field Grammar's getAttribute
    //    on "SubFieldNumber" - set if present
    _variant_t vSubFieldNumber(
        spFieldElementGrammar->getAttribute("SubFieldNumber"));
    if (vSubFieldNumber.vt != VT_NULL)
    {
      memset(cSubFieldNumber,0,sizeof(cSubFieldNumber));
      strcpy(cSubFieldNumber, _bstr_t(vSubFieldNumber));
      iSubFieldNumber = atoi(cSubFieldNumber);
      NewCell->setSubFieldNumber(iSubFieldNumber);
    }

    //  NewCell <- call NewCell's putField passing Element Text
    NewCell->putField(cElementText,strlen(cElementText));

    //  Get next field Element from Record Element
    spChildNode = spChildNode->nextSibling;
  } //  ENDDO

  return SUCCESS;
}

//  Process a group
int RecordWriter::processGroup()
{
  return FAILURE;
}

//  Write record to legacy format - stub for base class
int RecordWriter::writeRecord(IXMLDOMElementPtr spRecordGrammar)
{
  return FAILURE;
}

