/* SourceConverter.cpp

Base class for base SourceConverter class

From Chapter 6 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"


//  Special headers for this module only
#include <strstrea.h>


//  Default Constructor
SourceConverter::SourceConverter(bool boValidateIn, 
    const char * const cBaseOutputDirectoryIn) : Converter()
{
#include "BlasterExceptions.h"
  //  Set up implementation dependent environment for loading
  //  and parsing the FileDescription Document

  //  Locals
  //  COM
  int iPathLength;

  //  Validation Option <- Passed validation option
  boValidate = boValidateIn;

  //  Base Output Directory <- From passed base output
  //  directory, Appending directory separator character if
  //  required
  iPathLength = strlen(cBaseOutputDirectoryIn);

  if ( iPathLength > (MAX_PATH_LENGTH - 2) )
  {
    cerr << BUFFER_OVERFLOW_IN << 
      "SourceConverter::SourceConverter" << endl;
    throw cConstructorError;
  }

  memset(cBaseOutputDirectory,0,MAX_PATH_LENGTH);
  strcpy(cBaseOutputDirectory, cBaseOutputDirectoryIn);

  if ( cBaseOutputDirectory[iPathLength - 1] != '\\')
  {
    cBaseOutputDirectory[iPathLength] = '\\';
  }


  //  Perform any implementation specific setup for validating
  //     output documents before writing them
  //  Nothing to do for C++ and MSXML
}


//  Destructor
SourceConverter::~SourceConverter() {};


//  Stub functions for derived classes
int SourceConverter::processFile(
    const char * const cInputFileName)
{ return FAILURE; }

//  Process a group of records
int SourceConverter::processGroup(IXMLDOMNodePtr spParent,
      IXMLDOMElementPtr spGroupGrammar, int &iRecordLength)
{
  //  Locals
  //  XML
  IXMLDOMElementPtr spGroupElement;
  IXMLDOMElementPtr spRecordGrammar;
  IXMLDOMNodePtr spChildNode;
  IXMLDOMProcessingInstructionPtr spProlog;
  IXMLDOMAttributePtr spSchemaLocationAttribute;


  //  Basic
  char cGroupElementName[NMTOKEN_SIZE];
  char cGrammarElementName[NMTOKEN_SIZE];
  char cRecordElementName[NMTOKEN_SIZE];
  char cGrammarTagValue[CELL_BUFFER_SIZE];
  char cRecordTagValue[CELL_BUFFER_SIZE];
  int iReturn;

  //  Initializations
  memset(cGroupElementName,'\0',sizeof(cGroupElementName));
  memset(cGrammarElementName,'\0',sizeof(cGrammarElementName));
  memset(cRecordElementName,'\0',sizeof(cRecordElementName));

  //  Group Element Name <- Group Grammar
  //    getAttributeValue(ElementName)
  strcpy(cGroupElementName,
    _bstr_t(spGroupGrammar->getAttribute("ElementName")));

  //  Group Element <- createElement using Group Element Name
  spGroupElement = spDocOutput->createElement(cGroupElementName);

  //  Parent Element <- appendNode Group Element
  spParent->appendChild(spGroupElement);

  //  Grammar Element Name <- Record Grammar getNodeName
  strcpy(cGrammarElementName, spGroupGrammar->nodeName);

  //  IF Grammar Element Name = "Grammar"
  if (!(strcmp(cGrammarElementName,"Grammar")))
  {
     //  IF Schema Location URL is not NULL
    if (strlen(cSchemaLocationURL) > 0)
    {
      //  Create noNameSpaceSchemaLocation Attribute and
      //    append to Root Element
      _variant_t varType((short)NODE_ATTRIBUTE);
      spSchemaLocationAttribute = spDocOutput->createNode(
        varType,"xsi:noNamespaceSchemaLocation",
        "http://www.w3.org/2001/XMLSchema-instance");

      spGroupElement->setAttributeNode(spSchemaLocationAttribute);
      // We can finally set it now

      spGroupElement->setAttribute("xsi:noNamespaceSchemaLocation",
        cSchemaLocationURL);
    }  //    ENDIF
  }  //    ENDIF

  //  ChildNode <- Get Grammar Element's firstChild
  spChildNode = spGroupGrammar->firstChild;

  //  DO while ChildNode is not an Element Node
  while (spChildNode->nodeType != NODE_ELEMENT)
  {
    spChildNode = spChildNode->nextSibling;
  }
  
  //  Record Grammar Element <- ChildNode
  spRecordGrammar = spChildNode;


  //  Record Element Name <- Record Grammar
  //    getAttributeValue(ElementName)
  strcpy(cRecordElementName,
    _bstr_t(spRecordGrammar->getAttribute("ElementName")));

  //  Process the first record in the group
  //  Call RecordReader's parseRecord
  iReturn = Reader->parseRecord(spRecordGrammar);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  Call RecordReader's toXML
  iReturn = Reader->toXMLType();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  Call writeRecord
  Reader->writeRecord(spGroupElement, spRecordGrammar);

  //  RecordLength <- call RecordReader's readRecord
  iRecordLength = Reader->readRecord();

  // This advance in the grammar makes sure that we don't
  //    repeat the starting record of the group
  //  Record Grammar Element <- Get next Record Element from
  //    Group Grammar
  spChildNode = spChildNode->nextSibling;

  while ( (spChildNode != NULL ) &&
          (spChildNode->nodeType != NODE_ELEMENT) )
  {
    spChildNode = spChildNode->nextSibling;
  }

  if ( spChildNode == NULL )
  {
    // Only one record is defined for the group
    return SUCCESS;
  }

  spRecordGrammar = spChildNode;

  //  Grammar Record Tag <- Record Grammer getAttribute
  //    TagValue
  strcpy(cGrammarTagValue,
    _bstr_t(spRecordGrammar->getAttribute("TagValue")));

  //  DO until end of file
  while (iRecordLength >= 0)
  {
    //  Record Tag <- call RecordReader's  getRecordType
    memset(cRecordTagValue,0,sizeof(cRecordTagValue));
    iReturn = Reader->getRecordType(cRecordTagValue);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

      //  DO until Grammar Record Tag = Record Tag
    while (strcmp(cRecordTagValue,cGrammarTagValue))
    {
      //  Record Grammar Element <- Get next Record Element
      //    from Group Grammar
      spChildNode = spChildNode->nextSibling;

      //  Advance over comments, etc., if necessary
      while ( (spChildNode != NULL) &&
        (spChildNode->nodeType != NODE_ELEMENT) )
      {
        //  ChildNode <- nextSibling
        spChildNode = spChildNode->nextSibling;
      }

      //  IF Record Grammar Element is NULL
      if (spChildNode == NULL)
      {
        //  return - This record is not part of
        //    the group
        return SUCCESS;
      }  // ENDIF

      spRecordGrammar = spChildNode;

      //  Grammar Record Tag <- Record Grammer getAttribute
      //    TagValue
      memset(cGrammarTagValue,'\0',sizeof(cGrammarTagValue));
      strcpy(cGrammarTagValue,
        _bstr_t(spRecordGrammar->getAttribute("TagValue")));
    } // ENDDO


    //  Grammar Element Name <- Record Grammar getNodeName
    memset(cGrammarElementName,'\0',sizeof(cGrammarElementName));
    strcpy(cGrammarElementName, spRecordGrammar->nodeName);

    //  IF Grammar Element Name = "GroupDescription"
    if (!(strcmp(cGrammarElementName,"GroupDescription")))
    {
      //  Do recursive call of processGroup
      iReturn = processGroup(spGroupElement, spRecordGrammar,
        iRecordLength);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

    }
    else
    {
      //  Record Element Name <-
      //    Record Grammar getAttributeValue(ElementName)
      memset(cRecordElementName,'\0',sizeof(cRecordElementName));
      strcpy(cRecordElementName,
        _bstr_t(spRecordGrammar->getAttribute("ElementName")));

      //  Call RecordReader's parseRecord
      iReturn = Reader->parseRecord(spRecordGrammar);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

      //  Call RecordReader's toXML
      iReturn = Reader->toXMLType();
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

      //  Call writeRecord
      Reader->writeRecord(spGroupElement, spRecordGrammar);

      //  Input Record <- readRecord
      iRecordLength = Reader->readRecord();

    } //  ENDIF
  } //  ENDDO

  //  We should only get here on EOF
  return SUCCESS;
}


//  Save the output document
int SourceConverter::saveDocument(
    const IXMLDOMDocument2Ptr spDocOutput,
    const char * const cFileName,
    bool boValidate)
{
  //  Locals

  //  COM
  HRESULT hResult;

  //  XML
  IXMLDOMParseErrorPtr spParseError;

  //  Basic types

  //  Validate if requested
  if (boValidate)
  {
    spParseError = spDocOutput->validate();
    if( spParseError->errorCode != S_OK)
    {
      cerr << VALIDATION_ERROR_BRIEF << cFileName << endl;
      displayParseError(spParseError);
      return FAILURE;
    }
  }

  hResult = spDocOutput->save(cFileName);

  if FAILED(hResult)
  {
    cerr << XML_OUTPUT_SAVE_ERROR << cFileName << endl;
    return FAILURE;
  }

  return SUCCESS;
}
