/* TargetConverter.cpp

Base class for base TargetConverter class

From Chapter 6 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"


//  Special headers for this module only
#include <strstrea.h>


//  Default Constructor
TargetConverter::TargetConverter(const char * const
    cFileDescriptionDocumentName) : Converter()
{
#include "BlasterExceptions.h"

  int iReturn;
  //  Call loadFileDescriptionDocument from passed document
  iReturn = 
    loadFileDescriptionDocument(cFileDescriptionDocumentName);
  if (iReturn == FAILURE)
  {
    throw cConstructorError;
  }
}


//  Destructor
TargetConverter::~TargetConverter() {}

//  Stub functions for derived classes
int TargetConverter::processDocument(
     const IXMLDOMDocument2Ptr spDocInput) 
{ return FAILURE; }

//  Process a logical group of records in an XML Document
//  This method developed in Chapter 8
int TargetConverter::processGroup(IXMLDOMElementPtr spGroup,
                         IXMLDOMElementPtr spGroupGrammar)
{
  //  Locals
  //  XML
  IXMLDOMNodePtr spRecordGrammarNode;
  IXMLDOMElementPtr spRecordGrammarElement;
  IXMLDOMNodePtr spRecordNode;
  IXMLDOMElementPtr spRecordElement;
  char cRecordTag[CELL_BUFFER_SIZE];
  char cGrammarTag[NMTOKEN_SIZE];
  char cGrammarElementName[NMTOKEN_SIZE];
  int iReturn;

  //  Record Grammar Node <- Get firstChild from Grammar
  //    Element, skipping over non-Element Nodes
  spRecordGrammarNode = spGroupGrammar->firstChild;
  while (spRecordGrammarNode->nodeType != NODE_ELEMENT)
  {
    spRecordGrammarNode = spRecordGrammarNode->nextSibling;
  }

  //  Record Grammar Element <- Record Gramamr Node
  spRecordGrammarElement = spRecordGrammarNode;

  //  Process the Group's starting Record Element
  //  Record Element <- Get first childNode from Parent
  //    Element, skipping over non-Element Nodes
  spRecordNode = spGroup->firstChild;
  while (spRecordNode->nodeType != NODE_ELEMENT)
  {
    spRecordNode = spRecordNode->nextSibling;
  }
  spRecordElement = spRecordNode;

  //  Call RecordWriter's parseRecord for Record Element
  iReturn = Writer->parseRecord(spRecordElement,
      spRecordGrammarElement);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  Call class's writeRecord method
  iReturn = Writer->writeRecord(spRecordGrammarElement);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  // This advance makes sure that we don't repeat starting
  //    record in group
  //  Record Grammar Element <- Get next Record Element from
  //    Group Grammar, skipping over non-Element Nodes
  spRecordGrammarNode = spRecordGrammarNode->nextSibling;

  if ( spRecordGrammarNode == NULL)
  {
    return SUCCESS;   //  No more records defined in group
  }
  
  while (spRecordGrammarNode->nodeType != NODE_ELEMENT)
  {
    spRecordGrammarNode = spRecordGrammarNode->nextSibling;
    if ( spRecordGrammarNode == NULL)
    {
      return SUCCESS;   //  No more records defined in group
    }
  }

  spRecordGrammarElement = spRecordGrammarNode;
  //  Grammar Tag = call Record Grammar Element's getAttribute
  //    for "ElementName"
  memset(cGrammarTag,0,sizeof(cGrammarTag));
  strcpy(cGrammarTag,
      _bstr_t(spRecordGrammarElement->getAttribute("ElementName")));

  //  Record Elememt <- Get Record Element's nextSibling,
  //    skipping over non-Element Nodes
  spRecordNode = spRecordNode->nextSibling;

  //  DO until all child Elements of Parent have been processed
  while (spRecordNode != NULL)
  {
    if (spRecordNode->nodeType != NODE_ELEMENT)
    {
      spRecordNode = spRecordNode->nextSibling;
      continue;
    }
    spRecordElement = spRecordNode;
    //  Record Tag <- call Record Element's getNodeName
    memset(cRecordTag,0,sizeof(cRecordTag));
    strcpy(cRecordTag, _bstr_t(spRecordNode->nodeName));

    //  DO until Grammar Tag = Record Tag
    while (strcmp(cGrammarTag,cRecordTag))
    {
      //  Record Grammar Element <- Get next Record Element
      //    from Group Grammar, skipping over non-Element Nodes
      spRecordGrammarNode = spRecordGrammarNode->nextSibling;
      //  IF Record Grammar Element is NULL
      if (spRecordGrammarNode == NULL)
      {
        //  Return error  - This record is not part of the group
        cerr << cRecordTag << RECORD_NOT_IN_GROUP_GRAMMAR
         << _bstr_t(spGroupGrammar->getAttribute("ElementName"))
         << endl;
        return FAILURE;
      } //  ENDIF

      if (spRecordGrammarNode->nodeType != NODE_ELEMENT)
      {
        continue;
      }
      spRecordGrammarElement = spRecordGrammarNode;
      //  Grammar Tag = call Record Grammar Element's
      //    getAttribute for "ElementName"
      memset(cGrammarTag,0,sizeof(cGrammarTag));
      strcpy(cGrammarTag,
          _bstr_t(spRecordGrammarElement->getAttribute("ElementName")));
    } //  ENDDO

    //  Grammar Element Name <- Call Record Grammar Element's
    //    getNodeName
    memset(cGrammarElementName,0,sizeof(cGrammarElementName));
    strcpy(cGrammarElementName,
        _bstr_t(spRecordGrammarElement->nodeName));

    //  IF Grammar Element Name = "GroupDescription"
    if (!strcmp(cGrammarElementName,"GroupDescription"))
    {
      //  Call processGroup, passing Record Element and Record
      //    Grammar Element
      iReturn = processGroup(spRecordNode, spRecordGrammarElement);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }
    }
    else  //  ELSE
    {
      //  Call RecordWriter's parseRecord, passing Record
      //    Element and Record Grammar
      iReturn = Writer->parseRecord(spRecordNode, spRecordGrammarElement);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }
      //  Call class's writeRecord, passing Record Grammar
      //    Element
      iReturn = Writer->writeRecord(spRecordGrammarElement);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }
    } //  ENDIF

    //  Record Elememt <- Get Record Element's nextSibling,
    //    skipping over non-Element Nodes
    spRecordNode = spRecordNode->nextSibling;
  } //  ENDDO

  return SUCCESS;
}
